from importlib import import_module
from typing import Type

from pait.app.auto_load_app import auto_load_app_class
from pait.app.base.security.api_key import BaseAPIKey

pait_app_path: str = "pait.app." + auto_load_app_class().__name__.lower() + ".security.api_key"
APIKey: Type[BaseAPIKey] = getattr(import_module(pait_app_path), "APIKey")



import re
import string
from typing import Callable

from sanic import Blueprint, Sanic

from pait.app.base.simple_route import SimpleRoute, add_route_plugin
from pait.app.sanic.plugin.unified_response import UnifiedResponsePlugin


def replace_openapi_url_to_url(url: str) -> str:
    """Convert the OpenAPI URL format to a format supported by the web framework

    >>> assert "http://google.com/user/<user_id>/abc/<another_id>/def" == replace_openapi_url_to_url(
    >>>    "http://google.com/user/{user_id:int}/abc/{another_id:int}/def"
    >>> )
    """
    matches = re.findall(r"{([a-zA-Z_]+)}", url)
    for match in matches:
        url = url.replace("{" + match + "}", f"<{match}>")
    return url


def add_simple_route(
    app: Sanic,
    simple_route: "SimpleRoute",
    _replace_openapi_url_to_url: Callable[[str], str] = replace_openapi_url_to_url,
) -> None:
    add_route_plugin(simple_route, UnifiedResponsePlugin)
    app.add_route(simple_route.route, _replace_openapi_url_to_url(simple_route.url), methods=set(simple_route.methods))


def add_multi_simple_route(
    app: Sanic,
    *simple_route_list: "SimpleRoute",
    prefix: str = "/",
    title: str = "",
    _replace_openapi_url_to_url: Callable[[str], str] = replace_openapi_url_to_url,
) -> None:
    blueprint: Blueprint = Blueprint(
        title.translate(str.maketrans({key: "" for key in string.punctuation})).replace(" ", ""),  # type: ignore
        url_prefix=prefix,
    )
    for simple_route in simple_route_list:
        add_route_plugin(simple_route, UnifiedResponsePlugin)
        blueprint.add_route(
            simple_route.route, _replace_openapi_url_to_url(simple_route.url), methods=set(simple_route.methods)
        )
    app.blueprint(blueprint)



from pait.plugin.at_most_one_of import AtMostOneOfExtraParam, AtMostOneOfPlugin
from pait.plugin.required import RequiredExtraParam, RequiredGroupExtraParam, RequiredPlugin

from .check_json_resp import CheckJsonRespPlugin
from .mock_response import MockPlugin
from .unified_response import UnifiedResponsePlugin

__all__ = [
    "RequiredPlugin",
    "RequiredExtraParam",
    "RequiredGroupExtraParam",
    "AtMostOneOfPlugin",
    "AtMostOneOfExtraParam",
    "CheckJsonRespPlugin",
    "MockPlugin",
    "UnifiedResponsePlugin",
]



import sys
from tempfile import NamedTemporaryFile
from typing import IO, TYPE_CHECKING, Any, Callable, Dict, Generic, Optional, Type, TypeVar

from typing_extensions import Literal

from pait.model import response
from pait.plugin.base import PluginContext, PluginManager, PrePluginProtocol
from pait.util import get_pait_response_model

if TYPE_CHECKING:
    from pait.model.core import PaitCoreModel


RESP_T = TypeVar("RESP_T")


class MockPluginProtocol(PrePluginProtocol, Generic[RESP_T]):
    """Automatically return a json response with sample values based on the response object
    Note: the code logic of the routing function will not be executed
    """

    pait_response_model: Type[response.BaseResponseModel]
    example_column_name: Literal["example", "mock"]

    def __call__(self, context: PluginContext) -> Any:
        if self._is_async_func:
            return self.async_mock_response()
        return self.mock_response()

    @classmethod
    def pre_check_hook(cls, pait_core_model: "PaitCoreModel", kwargs: Dict) -> None:
        super().pre_check_hook(pait_core_model, kwargs)
        if not pait_core_model.response_model_list:
            raise RuntimeError(f"{pait_core_model.func} can not found response model")
        if "pait_response_model" in kwargs:
            raise RuntimeError("Please use response_model_list param")

    @classmethod
    def pre_load_hook(cls, pait_core_model: "PaitCoreModel", kwargs: Dict) -> Dict:
        kwargs = super().pre_load_hook(pait_core_model, kwargs)
        pait_response: Optional[Type[response.BaseResponseModel]] = None
        enable_mock_response_filter_fn: Optional[Callable] = kwargs.pop("enable_mock_response_filter_fn", None)
        if enable_mock_response_filter_fn and pait_core_model.response_model_list:
            for _pait_response in pait_core_model.response_model_list:
                if enable_mock_response_filter_fn(_pait_response):
                    pait_response = _pait_response
                    break

        if not pait_response:
            pait_response = get_pait_response_model(
                pait_core_model.response_model_list,
                target_pait_response_class=kwargs.pop("target_pait_response_class", False),
            )
        kwargs["pait_response_model"] = pait_response
        return kwargs

    def get_response(self) -> RESP_T:
        raise NotImplementedError()

    def get_file_response(self, temporary_file: IO[bytes], f: Any) -> RESP_T:
        raise RuntimeError("Not Implemented")  # pragma: no cover

    async def async_get_file_response(self, temporary_file: Any, f: Any) -> RESP_T:
        raise RuntimeError("Not Implemented")  # pragma: no cover

    def set_info_to_response(self, resp: RESP_T) -> None:
        raise NotImplementedError()

    def mock_response(self) -> RESP_T:
        if issubclass(self.pait_response_model, response.FileResponseModel):
            named_temporary_file: IO[bytes] = NamedTemporaryFile()
            f: Any = named_temporary_file.__enter__()
            try:
                resp: RESP_T = self.get_file_response(named_temporary_file, f)
            except Exception as e:
                exc_type, exc_val, exc_tb = sys.exc_info()
                named_temporary_file.__exit__(exc_type, exc_val, exc_tb)
                raise e
            self.set_info_to_response(resp)
        elif not issubclass(self.pait_response_model, response.BaseResponseModel):
            raise NotImplementedError(f"make_mock_response not support {self.pait_response_model}")
        else:
            resp = self.get_response()
        return resp

    async def async_mock_response(self) -> RESP_T:
        if issubclass(self.pait_response_model, response.FileResponseModel):
            import aiofiles  # type: ignore

            tf: aiofiles.tempfile.AsyncContextManager = aiofiles.tempfile.NamedTemporaryFile()  # type: ignore
            f: Any = await tf.__aenter__()
            try:
                resp: RESP_T = await self.async_get_file_response(tf, f)
            except Exception as e:
                exc_type, exc_val, exc_tb = sys.exc_info()
                await tf.__aexit__(exc_type, exc_val, exc_tb)
                raise e
            self.set_info_to_response(resp)
        elif not issubclass(self.pait_response_model, response.BaseResponseModel):
            raise NotImplementedError(f"make_mock_response not support {self.pait_response_model}")
        else:
            resp = self.get_response()
        return resp

    @classmethod
    def build(  # type: ignore
        cls,  # type: ignore
        enable_mock_response_filter_fn: Optional[Callable] = None,  # type: ignore
        target_pait_response_class: Optional[Type["response.BaseResponseModel"]] = None,  # type: ignore
        example_column_name: Literal["example", "mock"] = "example",  # type: ignore
    ) -> "PluginManager":  # type: ignore
        return super().build(
            enable_mock_response_filter_fn=enable_mock_response_filter_fn,
            target_pait_response_class=target_pait_response_class,
            example_column_name=example_column_name,
        )



from typing import TYPE_CHECKING, Any, Dict, List, Optional

from pydantic import Field

from pait.exceptions import CheckValueError
from pait.field import BaseField, ExtraParam
from pait.plugin.base import PluginContext, PluginManager, PostPluginProtocol
from pait.util import FuncSig, gen_tip_exc, get_func_sig

if TYPE_CHECKING:
    from pait.model.core import PaitCoreModel

__all__ = ["RequiredPlugin", "RequiredExtraParam", "RequiredGroupExtraParam"]


class RequiredGroupExtraParam(ExtraParam):
    group: str
    is_main: bool = Field(default=False)


class RequiredExtraParam(ExtraParam):
    main_column: str


class RequiredPlugin(PostPluginProtocol):
    """
    Check dependencies between parameters
    """

    required_dict: Dict[str, List[str]]

    def check_param(self, context: PluginContext) -> None:
        try:
            for pre_param, param_list in self.required_dict.items():
                if pre_param not in context.kwargs or not context.kwargs[pre_param]:
                    continue
                for param in param_list:
                    if context.kwargs.get(param, None) is None:
                        raise CheckValueError(
                            f"{pre_param} requires param {' and '.join(param_list)}, which if not none"
                        )
        except Exception as e:
            raise e from gen_tip_exc(context.pait_core_model.func, e)

    @classmethod
    def pre_load_hook(cls, pait_core_model: "PaitCoreModel", kwargs: Dict) -> Dict:
        kwargs = super().pre_load_hook(pait_core_model, kwargs)
        fun_sig: FuncSig = get_func_sig(pait_core_model.func)
        required_dict: Dict[str, List[str]] = {}
        temp_group_dict: Dict[str, List[str]] = {}
        group_main_dict: Dict[str, str] = {}
        for param in fun_sig.param_list:
            default: Any = param.default
            if not isinstance(default, BaseField):
                continue
            for extra_param in default.extra_param_list:
                column_name: str = default.alias or param.name
                if isinstance(extra_param, RequiredExtraParam):
                    required_dict.setdefault(extra_param.main_column, [])
                    required_dict[extra_param.main_column].append(column_name)
                elif isinstance(extra_param, RequiredGroupExtraParam):
                    if extra_param.is_main:
                        group_main_dict[extra_param.group] = column_name
                    else:
                        temp_group_dict.setdefault(extra_param.group, [])
                        temp_group_dict[extra_param.group].append(column_name)
        for group, column_name in group_main_dict.items():
            required_dict.setdefault(column_name, [])
            required_dict[column_name].extend(temp_group_dict[group])
        for k, v in required_dict.items():
            kwargs["required_dict"].setdefault(k, [])
            kwargs["required_dict"][k].extend(v)
        return kwargs

    @classmethod
    def build(cls, *, required_dict: Optional[Dict[str, List[str]]] = None) -> "PluginManager":  # type: ignore
        return super().build(required_dict=required_dict or {})

    def __call__(self, context: PluginContext) -> Any:
        self.check_param(context)
        return super().__call__(context)



from typing import IO, Any

import aiofiles  # type: ignore
from starlette.background import BackgroundTask
from starlette.responses import FileResponse, Response

from pait.app.starlette.adapter.response import gen_response, set_info_to_response
from pait.plugin.mock_response import RESP_T, MockPluginProtocol


class MockPlugin(MockPluginProtocol[Response]):
    def get_response(self) -> Response:
        return gen_response(
            self.pait_response_model.get_example_value(example_column_name=self.example_column_name),
            self.pait_response_model,
        )

    def get_file_response(self, temporary_file: IO[bytes], f: Any) -> RESP_T:  # type: ignore
        f.write(self.pait_response_model.get_example_value(example_column_name=self.example_column_name))
        f.seek(0)

        def close_file() -> None:
            temporary_file.__exit__(None, None, None)

        return FileResponse(
            f.name, media_type=self.pait_response_model.media_type, background=BackgroundTask(close_file)
        )

    async def async_get_file_response(self, temporary_file: Any, f: Any) -> RESP_T:
        await f.write(self.pait_response_model.get_example_value(example_column_name=self.example_column_name))
        await f.seek(0)

        async def close_file() -> None:
            await temporary_file.__aexit__(None, None, None)

        return FileResponse(
            f.name, media_type=self.pait_response_model.media_type, background=BackgroundTask(close_file)
        )

    def set_info_to_response(self, resp: Response) -> None:
        set_info_to_response(resp, self.pait_response_model)



import inspect
import json
import os
import sys
from datetime import date, datetime
from decimal import Decimal
from enum import Enum
from functools import wraps
from json import JSONEncoder
from typing import (  # type: ignore
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    ForwardRef,
    List,
    Optional,
    Set,
    Tuple,
    Type,
    TypeVar,
    Union,
    _eval_type,
    _GenericAlias,
    get_type_hints,
)

from pydantic import BaseModel
from pydantic.fields import FieldInfo, Undefined, UndefinedType
from typing_extensions import is_typeddict

from pait.field import BaseField, Depends, is_pait_field
from pait.types import ParamSpec

from ._types import ParseTypeError, parse_typing

if TYPE_CHECKING:
    from pait.model.response import BaseResponseModel

__all__ = [
    "gen_example_dict_from_pydantic_base_model",
    "gen_example_dict_from_schema",
    "gen_example_json_from_schema",
    "get_parameter_list_from_pydantic_basemodel",
    "gen_example_value_from_type",
    "get_pydantic_annotation",
    "get_parameter_list_from_class",
    "http_method_tuple",
    "json_type_default_value_dict",
    "python_type_default_value_dict",
    "get_pait_response_model",
    "example_value_handle",
    "ignore_pre_check",
    "gen_example_value_from_python",
    "get_real_annotation",
    "create_factory",
    "partial_wrapper",
    "R_T",
    "P",
]
ignore_pre_check: bool = bool(os.environ.get("PAIT_IGNORE_PRE_CHECK", False))
http_method_tuple: Tuple[str, ...] = ("get", "post", "head", "options", "delete", "put", "trace", "patch")

json_type_default_value_dict: Dict[str, Any] = {
    "null": None,
    "bool": True,
    "boolean": True,
    "string": "",
    "number": 0.0,
    "float": 0.0,
    "integer": 0,
    "object": {},
    "array": [],
}

python_type_default_value_dict: Dict = {
    bool: True,
    float: 0.0,
    int: 0,
    str: "",
    list: [],
    tuple: (),
    dict: {},
    datetime: datetime.fromtimestamp(0),
    date: date.today(),
    Decimal: Decimal("0.0"),
    Any: {},
}


P = ParamSpec("P")
R_T = TypeVar("R_T")


def create_factory(func: Callable[P, R_T]) -> Callable[P, Callable[[], R_T]]:
    """Create a factory that calls the function (Use the syntax hints provided by PEP 612)"""

    @wraps(func)
    def wrapper(*args: P.args, **kwargs: P.kwargs) -> Callable[[], R_T]:
        return lambda: func(*args, **kwargs)

    return wrapper


def partial_wrapper(func: Callable[P, R_T], **_customer_kwargs: Any) -> Callable[P, R_T]:  # type: ignore
    """with type hints partial"""

    @wraps(func)
    def wrapper(*args: P.args, **kwargs: P.kwargs) -> R_T:  # type: ignore
        new_kwargs = _customer_kwargs.copy()
        new_kwargs.update(kwargs)
        return func(*args, **new_kwargs)

    return wrapper


def example_value_handle(example_value: Any) -> Any:
    if isinstance(example_value, Enum):
        example_value = example_value.value
    elif getattr(example_value, "__call__", None):
        example_value = example_value()
    return example_value


def get_real_annotation(annotation: Union[Type, str], target_obj: Any) -> Type:
    """
    get the real annotation from postponed annotations/annotations
    :param annotation: type hints
    :param target_obj: The object on which the annotation resides
    :return: the real annotation

    e.g:
    >>> def demo(a: "int") -> int: pass
    >>> assert int = get_real_annotation(demo.__annotations__["a"], demo)

    >>> class Demo:
    >>>     a: "int"
    >>> assert int = get_real_annotation(Demo.__annotations__["a"], Demo)
    """
    global_dict = sys.modules[target_obj.__module__].__dict__
    if not isinstance(annotation, str):
        # Option["Dict[str, int]"]
        return _eval_type(annotation, global_dict, None)
    if inspect.isclass(target_obj) and annotation in target_obj.__dict__:
        new_annotation: Type = target_obj.__dict__[annotation]
    else:
        # get real type
        value: ForwardRef = ForwardRef(annotation, is_argument=False)

        if sys.version_info >= (3, 9):
            new_annotation = value._evaluate(global_dict, None, frozenset())  # type: ignore    # pragma: no cover
        else:
            new_annotation = value._evaluate(global_dict, None)  # type: ignore
        if not new_annotation:
            raise RuntimeError(f"get real annotation from {target_obj} fail")  # pragma: no cover
    return _eval_type(new_annotation, global_dict, None)


def get_pydantic_annotation(key: str, pydantic_base_model: Type[BaseModel]) -> Type:
    """Get the annotation from BaseModel's properties"""
    annotation = pydantic_base_model.__fields__[key].annotation
    annotation = get_real_annotation(annotation, pydantic_base_model)
    if getattr(annotation, "real", None) and annotation != bool:
        # support like pydantic.ConstrainedIntValue
        return annotation.real.__objclass__  # type: ignore
    __base__ = getattr(annotation, "__base__", None)
    if __base__ and getattr(__base__, "__module__", "") == "pydantic.types":
        return annotation.__mro__[2]
    return annotation


def get_pait_response_model(
    response_model_list: List[Type["BaseResponseModel"]],
    target_pait_response_class: Optional[Type["BaseResponseModel"]] = None,
) -> Type["BaseResponseModel"]:
    if target_pait_response_class:
        response_model_list = [i for i in response_model_list if issubclass(i, target_pait_response_class)]
    return response_model_list[0]


def gen_example_value_from_python(obj: Any) -> Any:
    if isinstance(obj, dict):
        return {key: gen_example_value_from_python(value) for key, value in obj.items()}
    else:
        return python_type_default_value_dict.get(type(obj), obj)


def gen_example_value_from_type(value_type: type, example_column_name: str = "example") -> Any:
    """
    Gets the default value for type
    :param value_type: type of the value
    :param example_column_name: Gets sample values from the properties specified by pydantic.FieldInfo
    """
    if is_typeddict(value_type):
        return {k: gen_example_value_from_type(v) for k, v in value_type.__annotations__.items()}
    elif isinstance(value_type, _GenericAlias):
        sub_type: Optional[Type] = None
        try:
            parse_typing_result = parse_typing(value_type)
            # If there is more than one type value, only the first one is used
            real_type: Type = parse_typing_result[0]
            annotation_arg_list: list = getattr(value_type, "__args__", [])
            if annotation_arg_list:
                sub_type_set: Set[Type] = set(annotation_arg_list)
                if len(sub_type_set) == 1:
                    sub_type = sub_type_set.pop()
        except ParseTypeError:
            real_type = value_type  # pragma: no cover
        if real_type is list and sub_type:
            return [gen_example_value_from_type(sub_type, example_column_name=example_column_name)]
        else:
            # if real_type is Option[xxx], sub_type is None
            # so must parse real_type
            return gen_example_value_from_type(real_type, example_column_name=example_column_name)
    elif not inspect.isclass(value_type):
        return python_type_default_value_dict[value_type]  # pragma: no cover
    elif issubclass(value_type, Enum):
        return [i for i in value_type.__members__.values()][0].value
    elif issubclass(value_type, BaseModel):
        return gen_example_dict_from_pydantic_base_model(value_type, example_column_name=example_column_name)
    else:
        return python_type_default_value_dict[value_type]


def gen_example_dict_from_pydantic_base_model(
    pydantic_base_model: Type[BaseModel], example_column_name: str = "example"
) -> dict:
    """
    Gets the default value for pydantic.BaseModel
    :param pydantic_base_model: pydantic.BaseModel
    :param example_column_name: Gets sample values from the properties specified by pydantic.FieldInfo
    """
    gen_dict: Dict[str, Any] = {}
    for key, model_field in pydantic_base_model.__fields__.items():
        if model_field.alias:
            real_key = model_field.alias
        if example_column_name:
            example_value: Any = model_field.field_info.extra.get(example_column_name, Undefined)
            if not isinstance(example_value, UndefinedType):
                gen_dict[real_key] = example_value_handle(example_value)
                continue

        if not isinstance(model_field.field_info.default, UndefinedType):
            gen_dict[real_key] = model_field.field_info.default
        elif model_field.field_info.default_factory and not isinstance(
            model_field.field_info.default_factory, UndefinedType
        ):
            gen_dict[real_key] = model_field.field_info.default_factory()
        else:
            annotation: Type = get_pydantic_annotation(key, pydantic_base_model)
            gen_dict[real_key] = gen_example_value_from_type(annotation, example_column_name=example_column_name)
    return gen_dict


def gen_example_dict_from_schema(schema_dict: Dict[str, Any], definition_dict: Optional[dict] = None) -> Dict[str, Any]:
    gen_dict: Dict[str, Any] = {}
    if "enum" in schema_dict:
        return schema_dict["enum"][0]
    elif "properties" not in schema_dict:
        return gen_dict
    property_dict: Dict[str, Any] = schema_dict["properties"]
    if not definition_dict:
        _definition_dict: dict = schema_dict.get("definitions", {})
    else:
        _definition_dict = definition_dict
    for key, value in property_dict.items():
        if "items" in value and value["type"] == "array":
            if "$ref" in value["items"]:
                model_key: str = value["items"]["$ref"].split("/")[-1]
                gen_dict[key] = [gen_example_dict_from_schema(_definition_dict.get(model_key, {}), _definition_dict)]
            else:
                gen_dict[key] = []
        elif "$ref" in value:
            model_key = value["$ref"].split("/")[-1]
            gen_dict[key] = gen_example_dict_from_schema(_definition_dict.get(model_key, {}), _definition_dict)
        else:
            if "example" in value:
                gen_dict[key] = example_value_handle(value["example"])
            elif "default" in value:
                gen_dict[key] = value["default"]
            else:
                if "type" in value:
                    if value["type"] not in json_type_default_value_dict:
                        raise KeyError(f"Can not found type: {key} in json type")
                    gen_dict[key] = json_type_default_value_dict[value["type"]]
                else:
                    gen_dict[key] = "object"
            # if isinstance(gen_dict[key], Enum):
            #     gen_dict[key] = gen_dict[key].value
    return gen_dict


def gen_example_json_from_schema(schema_dict: Dict[str, Any], cls: Optional[Type[JSONEncoder]] = None) -> str:
    return json.dumps(gen_example_dict_from_schema(schema_dict), cls=cls)


def get_parameter_list_from_pydantic_basemodel(
    pait_model: Type[BaseModel], default_field_class: Optional[Type[BaseField]] = None
) -> List["inspect.Parameter"]:
    """get class parameter list by attributes, if attributes not default value, it will be set `Undefined`"""
    key = f"_parameter_list:{default_field_class}"
    parameter_list: Optional[List["inspect.Parameter"]] = getattr(pait_model, key, None)
    if parameter_list is not None:
        return parameter_list
    parameter_list = []
    for key, model_field in pait_model.__fields__.items():
        field: FieldInfo = model_field.field_info
        if not is_pait_field(field):
            if not default_field_class:
                raise TypeError(  # pragma: no cover
                    f"{field.__class__} must instance {BaseField} or {Depends} by model {pait_model}"
                )
            field = default_field_class.from_pydantic_field(field)

            if getattr(field, "alias", None) is None:
                field.request_key = key
        parameter = inspect.Parameter(
            key,
            inspect.Parameter.POSITIONAL_ONLY,
            default=field,
            annotation=get_pydantic_annotation(key, pait_model),
        )
        parameter_list.append(parameter)

    setattr(pait_model, key, parameter_list)
    return parameter_list


_class_parameter_list_dict: Dict[type, List["inspect.Parameter"]] = {}


def get_parameter_list_from_class(cbv_class: Type) -> List["inspect.Parameter"]:
    """get class parameter list by attributes, if attributes not default value, it will be set `Undefined`"""
    parameter_list: Optional[List["inspect.Parameter"]] = _class_parameter_list_dict.get(cbv_class)
    if parameter_list is not None:
        return parameter_list
    parameter_list = []
    if hasattr(cbv_class, "__annotations__"):
        for param_name, param_annotation in get_type_hints(cbv_class).items():
            default: Any = getattr(cbv_class, param_name, Undefined)
            if not is_pait_field(default):
                continue

            # Optimize parsing speed
            if getattr(default, "alias", None) is None:
                default.request_key = param_name
            parameter: "inspect.Parameter" = inspect.Parameter(
                param_name,
                inspect.Parameter.POSITIONAL_ONLY,
                default=default,
                annotation=param_annotation,
            )
            parameter_list.append(parameter)
    _class_parameter_list_dict[cbv_class] = parameter_list
    return parameter_list



import inspect
import json
import logging
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Type, Union

from any_api.openapi import ApiModel as _ApiModel
from any_api.openapi import ExternalDocumentationModel, HttpParamTypeLiteral, InfoModel
from any_api.openapi import OpenAPI as _OpenAPI
from any_api.openapi import SecurityModelType, ServerModel, TagModel, openapi_model
from any_api.openapi.model.links import LinksModel
from any_api.openapi.model.openapi import OpenAPIModel
from any_api.openapi.model.requests import RequestModel
from pydantic import BaseModel, Field
from pydantic.fields import FieldInfo, Undefined

from pait.app import load_app
from pait.field import BaseField, Depends
from pait.g import config
from pait.model.core import PaitCoreModel
from pait.types import CallType
from pait.util import FuncSig, create_pydantic_model, get_func_sig, get_parameter_list_from_class

HttpParamTypeDictType = Dict[HttpParamTypeLiteral, List[RequestModel]]


__all__ = ["LinksModel", "ApiModel", "ParsePaitModel", "OpenAPI"]


class ApiModel(_ApiModel):
    pait_core_model: PaitCoreModel = Field()

    class Config:
        arbitrary_types_allowed = True

    def add_to_operation_model(self, _openapi_model: openapi_model.OperationModel) -> None:
        # Not an openapi standard parameter
        _openapi_model.pait_info = {
            "group": self.pait_core_model.group,
            "status": self.pait_core_model.status.value,
            "author": self.pait_core_model.author,
            "pait_id": self.pait_core_model.pait_id,
        }


class ParsePaitModel(object):
    def __init__(self, pait_model: PaitCoreModel) -> None:
        self.pait_model: PaitCoreModel = pait_model
        self.http_param_type_dict: HttpParamTypeDictType = {}
        self.security_dict: Dict[str, openapi_model.security.SecurityModelType] = {}
        self.http_param_type_alias_dict: Dict[str, HttpParamTypeLiteral] = {"multiquery": "query"}

        self.param_field_dict: Dict[str, BaseField] = {}
        self.http_param_type_annotation_dict: Dict[HttpParamTypeLiteral, Dict[str, Tuple[Type, FieldInfo]]] = {}

        for extra_openapi_model in self.pait_model.extra_openapi_model_list:
            self._parse_base_model(extra_openapi_model)
        for pre_depend in pait_model.pre_depend_list:
            self._parse_call_type(pre_depend)
        self._parse_call_type(pait_model.func)

        for http_param_type, annotation_dict in self.http_param_type_annotation_dict.items():
            http_param_type = self.http_param_type_alias_dict.get(http_param_type, http_param_type)
            if http_param_type not in self.http_param_type_dict:
                self.http_param_type_dict[http_param_type] = []
            self.http_param_type_dict[http_param_type].append(
                RequestModel(
                    description="",
                    media_type_list=[self.param_field_dict[http_param_type].media_type],
                    openapi_serialization=self.param_field_dict[http_param_type].openapi_serialization,
                    model=create_pydantic_model(
                        annotation_dict,
                        class_name=f"{self.pait_model.func_name.title()}{self.pait_model.pait_id.title()}",
                    ),
                )
            )

    def _parse_base_model(
        self, _pydantic_model: Type[BaseModel], default_field_class: Optional[Type[BaseField]] = None
    ) -> None:
        from typing import get_type_hints

        for field_name, model_field in _pydantic_model.__fields__.items():
            param_annotation = get_type_hints(_pydantic_model)[field_name]
            field = model_field.field_info
            if not isinstance(field, BaseField):
                if self.pait_model.default_field_class:
                    field = self.pait_model.default_field_class.from_pydantic_field(field)
                elif default_field_class:
                    field = default_field_class.from_pydantic_field(field)
                else:
                    continue
            if not field.openapi_include:
                continue
            if isinstance(field, BaseField) and field.alias:
                field_name = field.alias
            http_param_type: HttpParamTypeLiteral = field.get_field_name()  # type: ignore
            http_param_type = self.http_param_type_alias_dict.get(http_param_type, http_param_type)
            if http_param_type not in self.http_param_type_annotation_dict:
                self.http_param_type_annotation_dict[http_param_type] = {}
            self.http_param_type_annotation_dict[http_param_type][field_name] = (param_annotation, field)
            self.param_field_dict[http_param_type] = field

    def parameter_list_handle(
        self,
        parameter_list: List["inspect.Parameter"],
        single_field_list: List[Tuple[str, "inspect.Parameter"]],
    ) -> None:
        """parse parameter_list to field_dict and single_field_list"""
        for parameter in parameter_list:
            if parameter.default != parameter.empty:
                annotation: type = parameter.annotation
                pait_field: Union[BaseField, Depends] = parameter.default
                if (
                    inspect.isclass(annotation)
                    and issubclass(annotation, BaseModel)
                    and not isinstance(pait_field, Depends)
                ):
                    # support def test(pait_model_route: BaseModel = Body())
                    if not pait_field.openapi_include:
                        continue
                    if not pait_field.raw_return:
                        self._parse_base_model(
                            create_pydantic_model({parameter.name: (parameter.annotation, pait_field)}),
                            pait_field.__class__,
                        )
                    else:
                        self._parse_base_model(annotation, pait_field.__class__)
                else:
                    # def test(pait_model_route: int = Body())
                    if isinstance(pait_field, Depends):
                        self._parse_call_type(pait_field.func)
                    else:
                        if not pait_field.openapi_include:
                            continue
                        field_name: str = pait_field.__class__.__name__.lower()
                        single_field_list.append((field_name, parameter))

            elif inspect.isclass(parameter.annotation) and issubclass(parameter.annotation, BaseModel):
                # def test(pait_model_route: PaitBaseModel)
                self._parse_base_model(parameter.annotation)

    def _parse_call_type(self, call_type: CallType) -> None:
        from pait.app.base.security.base import BaseSecurity

        if isinstance(call_type, BaseSecurity):
            if self.security_dict.get(call_type.security_name, None) not in (call_type.model, None):
                raise ValueError(
                    f"{call_type.security_name}Already exists, "
                    f"but the Security Model is inconsistent with {call_type.model}"
                )
            self.security_dict[call_type.security_name] = call_type.model

        func_type_sig: FuncSig = get_func_sig(call_type)
        single_field_list: List[Tuple[str, "inspect.Parameter"]] = []

        qualname: str = getattr(call_type, "__qualname__", "")
        if not qualname:
            class_ = call_type.__class__  # type: ignore
        else:
            qualname = qualname.split(".<locals>", 1)[0].rsplit(".", 1)[0]
            class_ = getattr(inspect.getmodule(call_type), qualname)

        if inspect.isclass(class_):
            parameter_list: List["inspect.Parameter"] = get_parameter_list_from_class(class_)
            self.parameter_list_handle(parameter_list, single_field_list)
        self.parameter_list_handle(func_type_sig.param_list, single_field_list)

        if single_field_list:
            annotation_dict: Dict[str, Tuple[Type, Any]] = {}
            _column_name_set: Set[str] = set()
            for field_name, parameter in single_field_list:
                field: BaseField = parameter.default
                key: str = field.alias or parameter.name
                if key in _column_name_set:
                    # Since the same name cannot exist together in a Dict,
                    #  it will be parsed directly when a Key exists
                    # fix
                    #  class Demo(BaseModel):
                    #      header_token: str = Header(alias="token")
                    #      query_token: str = Query(alias="token")
                    _pydantic_model: Type[BaseModel] = create_pydantic_model(
                        {parameter.name: (parameter.annotation, field)},
                        class_name=(
                            f"{self.pait_model.func_name.title()}"
                            f"{parameter.name.title()}{self.pait_model.pait_id.title()}"
                        ),
                    )
                    self._parse_base_model(_pydantic_model)
                else:
                    _column_name_set.add(key)
                    annotation_dict[parameter.name] = (parameter.annotation, field)

            _pydantic_model = create_pydantic_model(
                annotation_dict, class_name=f"{self.pait_model.func_name.title()}{self.pait_model.pait_id.title()}"
            )
            self._parse_base_model(_pydantic_model)


class OpenAPI(object):
    def __init__(
        self,
        app: Any,
        undefined: Any = Undefined,
        openapi_info_model: Optional[InfoModel] = None,
        server_model_list: Optional[List[ServerModel]] = None,
        tag_model_list: Optional[List[TagModel]] = None,
        external_docs: Optional[ExternalDocumentationModel] = None,
        security_dict: Optional[Dict[str, SecurityModelType]] = None,
    ):
        self._undefined: Any = undefined
        self._openapi: _OpenAPI = _OpenAPI(
            openapi_info_model=openapi_info_model,
            server_model_list=server_model_list,
            tag_model_list=tag_model_list,
            external_docs=external_docs,
            security_dict=security_dict,
        )
        self._pait_dict: Dict[str, PaitCoreModel] = load_app(app)
        api_model_list: List[ApiModel] = []
        for pait_id, pait_model in self._pait_dict.items():
            try:
                parse_pait_model: ParsePaitModel = ParsePaitModel(pait_model)
                api_model_list.append(
                    ApiModel(
                        path=pait_model.openapi_path,
                        http_method_list=pait_model.openapi_method_list,
                        tags=[i.to_tag_model() for i in pait_model.tag],
                        operation_id=pait_model.operation_id,
                        summary=pait_model.summary,
                        request_dict=parse_pait_model.http_param_type_dict,
                        response_list=pait_model.response_model_list,
                        description=pait_model.desc,
                        deprecated=pait_model.status.is_deprecated(),
                        pait_core_model=pait_model,
                        security=parse_pait_model.security_dict,
                    )
                )
            except Exception as e:
                logging.error(f"parse pait model error by func: {pait_model.func_path}/{pait_model.func_name}")
                raise e
        # In order to be compatible with the link, it must be imported in batches
        self._openapi.add_api_model(*api_model_list)

    #########################
    # proxy openapi feature #
    #########################
    @property
    def model(self) -> OpenAPIModel:
        return self._openapi.model

    @property
    def dict(self) -> dict:
        return self._openapi.dict

    def content(self, serialization_callback: Callable = json.dumps, **kwargs: Any) -> str:
        if serialization_callback is json.dumps and "cls" not in kwargs:
            kwargs["cls"] = config.json_encoder
        return self._openapi.content(serialization_callback, **kwargs)



import copy
from typing import Any, Dict, Optional, Type

from any_api.openapi.model.responses import BaseResponseModel
from any_api.openapi.model.responses import FileResponseModel as _FileResponseModel
from any_api.openapi.model.responses import HtmlResponseModel as _HtmlResponseModel
from any_api.openapi.model.responses import JsonResponseModel as _JsonResponseModel
from any_api.openapi.model.responses import TextResponseModel as _TextResponseModel
from any_api.openapi.model.responses import XmlResponseModel as _XmlResponseModel
from pydantic import BaseModel

from pait.util import gen_example_dict_from_pydantic_base_model

__all__ = [
    "ResponseModel",
    "PaitResponseModel",
    "BaseResponseModel",
    "HtmlResponseModel",
    "HtmlResponseModel",
    "JsonResponseModel",
    "JsonResponseModel",
    "FileResponseModel",
    "FileResponseModel",
    "TextResponseModel",
    "TextResponseModel",
    "XmlResponseModel",
    "XmlResponseModel",
    "HttpStatusCodeBaseModel",
    "Http400RespModel",
    "Http401RespModel",
    "Http403RespModel",
    "Http404RespModel",
    "Http405RespModel",
    "Http406RespModel",
    "Http407RespModel",
    "Http408RespModel",
    "Http429RespModel",
]


class JsonResponseModel(_JsonResponseModel):
    @classmethod
    def _get_example_dict(cls, model: Type[BaseModel], **kwargs: Any) -> dict:
        return gen_example_dict_from_pydantic_base_model(model, **kwargs)

    @classmethod
    def get_default_dict(cls, **extra: Any) -> dict:
        default_dict: dict = getattr(cls.response_data, "JsonResponseModel_default_dict", {})
        if not default_dict:
            default_dict = gen_example_dict_from_pydantic_base_model(
                cls.response_data, example_column_name=extra.pop("example_column_name", "")
            )
            setattr(cls.response_data, "JsonResponseModel_default_dict", default_dict)
        return copy.deepcopy(default_dict)


class XmlResponseModel(_XmlResponseModel):
    @classmethod
    def _get_example_dict(cls, model: Type[BaseModel], **kwargs: Any) -> dict:
        return gen_example_dict_from_pydantic_base_model(model, **kwargs)

    @classmethod
    def get_default_dict(cls, **extra: Any) -> dict:
        default_dict: dict = getattr(cls.response_data, "XmlResponseModel_default_dict", {})
        if not default_dict:
            default_dict = gen_example_dict_from_pydantic_base_model(
                cls.response_data, example_column_name=extra.pop("example_column_name", "")
            )
            setattr(cls.response_data, "XmlJsonResponseModel_default_dict", default_dict)
        return copy.deepcopy(default_dict)


class TextResponseModel(_TextResponseModel):
    @classmethod
    def _get_example_dict(cls, model: Type[BaseModel], **extra: Any) -> dict:
        return gen_example_dict_from_pydantic_base_model(model, **extra)


class HtmlResponseModel(_HtmlResponseModel):
    @classmethod
    def _get_example_dict(cls, model: Type[BaseModel], **extra: Any) -> dict:
        return gen_example_dict_from_pydantic_base_model(model, **extra)


class FileResponseModel(_FileResponseModel):
    @classmethod
    def _get_example_dict(cls, model: Type[BaseModel], **extra: Any) -> dict:
        return gen_example_dict_from_pydantic_base_model(model, **extra)


###################################
# Compatible with old version API #
###################################
class PaitJsonResponseModel(JsonResponseModel):
    pass


class ResponseModel(JsonResponseModel):
    pass


class PaitResponseModel(JsonResponseModel):
    """JsonResponseModel alias Compatible versions below 0.7"""


class PaitTextResponseModel(TextResponseModel):
    pass


class PaitHtmlResponseModel(HtmlResponseModel):
    pass


class PaitXmlResponseModel(XmlResponseModel):
    pass


class PaitFileResponseModel(FileResponseModel):
    pass


# Only define the status code that Pait will use internally, the user needs to define it through config.init config
http_status_code_dict: Dict[int, str] = {
    400: "Bad Request",
    401: "Unauthorized",
    403: "Forbidden",
    404: "Not Found",
    405: "Method Not Allowed",
    406: "Not Accepted",
    407: "Proxy Authentication Required",
    408: "Request Timeout",
    429: "Too Many Requests",
}


class HttpStatusCodeBaseModel(BaseResponseModel):
    response_data = None

    @classmethod
    def clone(
        cls, resp_model: Type[BaseResponseModel], status_code: Optional[int] = None, response_data: Any = None
    ) -> "Type[HttpStatusCodeBaseModel]":
        status_code = status_code or cls.status_code[0]
        response_data = response_data or http_status_code_dict.get(status_code, "") or cls.response_data
        if issubclass(resp_model, HtmlResponseModel):
            response_data = f"<h1>{response_data}<h1/>"
        resp_model_type = resp_model.__class__(  # type: ignore
            f"Http{status_code}RespModel",
            (
                resp_model,
                cls,
            ),
            {}
            # {"clone": cls.clone},
        )
        resp_model_type.status_code = (status_code,)
        resp_model_type.response_data = response_data
        return resp_model_type


Http400RespModel = HttpStatusCodeBaseModel.clone(resp_model=HtmlResponseModel, status_code=400)
Http401RespModel = HttpStatusCodeBaseModel.clone(resp_model=HtmlResponseModel, status_code=401)
Http403RespModel = HttpStatusCodeBaseModel.clone(resp_model=HtmlResponseModel, status_code=403)
Http404RespModel = HttpStatusCodeBaseModel.clone(resp_model=HtmlResponseModel, status_code=404)
Http405RespModel = HttpStatusCodeBaseModel.clone(resp_model=HtmlResponseModel, status_code=405)
Http406RespModel = HttpStatusCodeBaseModel.clone(resp_model=HtmlResponseModel, status_code=406)
Http407RespModel = HttpStatusCodeBaseModel.clone(resp_model=HtmlResponseModel, status_code=407)
Http408RespModel = HttpStatusCodeBaseModel.clone(resp_model=HtmlResponseModel, status_code=408)
Http429RespModel = HttpStatusCodeBaseModel.clone(resp_model=HtmlResponseModel, status_code=429)



from typing import Any, Generic, List, Mapping, TypeVar

RequestT = TypeVar("RequestT")


class BaseRequestExtend(Generic[RequestT]):
    def __init__(self, request: RequestT) -> None:
        self.request: RequestT = request

    @property
    def scheme(self) -> str:
        raise NotImplementedError

    @property
    def path(self) -> str:
        raise NotImplementedError

    @property
    def hostname(self) -> str:
        raise NotImplementedError


RequestExtendT = TypeVar("RequestExtendT", bound=BaseRequestExtend)


class BaseRequest(Generic[RequestT, RequestExtendT]):
    # The class that defines the request object corresponding to the framework (consistent with Request T)
    RequestType: RequestT = type(None)  # type: ignore
    FormType = type(None)  # The class that defines the form object corresponding to the framework
    FileType = type(None)  # The class that defines the file object corresponding to the framework
    HeaderType = type(None)  # The class that defines the header object corresponding to the framework

    def __init__(self, request: RequestT, args: List[Any], kwargs: Mapping[str, Any]):
        self.request: RequestT = request
        self.args: List[Any] = args
        self.kwargs: Mapping[str, Any] = kwargs
        self.request_kwargs: Mapping[str, Any] = kwargs

    def request_extend(self) -> BaseRequestExtend[RequestT]:
        return BaseRequestExtend(self.request)

    def cookie(self) -> Any:
        raise NotImplementedError

    def header(self) -> Any:
        raise NotImplementedError

    def path(self) -> Any:
        raise NotImplementedError

    def query(self) -> Any:
        raise NotImplementedError

    def json(self) -> dict:
        return self.body()

    def body(self) -> Any:
        raise NotImplementedError

    def file(self) -> Any:
        raise NotImplementedError

    def form(self) -> Any:
        raise NotImplementedError

    def multiform(self) -> Any:
        raise NotImplementedError

    def multiquery(self) -> Any:
        raise NotImplementedError

    def check_request_type(self, value: Any) -> bool:
        return value is self.RequestType

    def check_file_type(self, value: Any) -> bool:
        return value is self.FileType

    def check_form_type(self, value: Any) -> bool:
        return value is self.FormType

    def check_header_type(self, value: Any) -> bool:
        return value is self.HeaderType



from typing import Optional

from tornado.web import HTTPError


def http_exception(*, status_code: int, message: str, headers: Optional[dict] = None) -> Exception:
    exc: HTTPError = HTTPError(reason=message, status_code=status_code)
    if headers:
        exc.headers = headers
    return exc



from importlib import import_module

from pait.app.auto_load_app import auto_load_app_class
from pait.plugin.auto_complete_json_resp import AutoCompleteJsonRespPlugin as _AutoCompleteJsonRespPlugin

__all__ = ["AutoCompleteJsonRespPlugin"]
pait_app_path: str = "pait.app." + auto_load_app_class().__name__.lower()
AutoCompleteJsonRespPlugin: "_AutoCompleteJsonRespPlugin" = getattr(
    import_module(pait_app_path + ".plugin"), "AutoCompleteJsonRespPlugin"
)



from pait.app.base.security.http import BaseHTTPBasic, BaseHTTPBearer, BaseHTTPDigest, HTTPBasicCredentials

from .util import GetException

__all__ = ["HTTPBasic", "HTTPDigest", "HTTPBearer", "HTTPBasicCredentials"]


class HTTPBasic(GetException, BaseHTTPBasic):
    pass


class HTTPDigest(GetException, BaseHTTPDigest):
    pass


class HTTPBearer(GetException, BaseHTTPBearer):
    pass



from typing import TYPE_CHECKING, Any, Dict, Optional, Tuple, Type

from pydantic import create_model

if TYPE_CHECKING:
    from pydantic import BaseConfig, BaseModel


__all__ = ["create_pydantic_model"]


def create_pydantic_model(
    annotation_dict: Optional[Dict[str, Tuple[Type, Any]]] = None,
    class_name: str = "DynamicModel",
    pydantic_config: Optional[Type["BaseConfig"]] = None,
    pydantic_base: Optional[Type["BaseModel"]] = None,
    pydantic_module: str = "pydantic.main",
    pydantic_validators: Optional[Dict[str, classmethod]] = None,
) -> Type["BaseModel"]:
    """pydantic self.pait_response_model helper
    if use create_model('DynamicModel', **annotation_dict), mypy will tip error
    """
    return create_model(
        class_name,
        __config__=pydantic_config,
        __base__=pydantic_base,
        __module__=pydantic_module,
        __validators__=pydantic_validators,
        **(annotation_dict or {}),
    )



from importlib import import_module
from typing import Type

from pait.app.auto_load_app import auto_load_app_class
from pait.app.base.security.oauth2 import (
    BaseOAuth2PasswordBearer,
    BaseOAuth2PasswordRequestFrom,
    OAuth2PasswordRequestFrom,
    OAuth2PasswordRequestFromStrict,
)

__all__ = [
    "OAuth2PasswordBearer",
    "OAuth2PasswordRequestFrom",
    "OAuth2PasswordRequestFromStrict",
    "BaseOAuth2PasswordRequestFrom",
]

pait_app_path: str = "pait.app." + auto_load_app_class().__name__.lower() + ".security.oauth2"
OAuth2PasswordBearer: Type[BaseOAuth2PasswordBearer] = getattr(import_module(pait_app_path), "OAuth2PasswordBearer")



from dataclasses import MISSING
from typing import Any

from starlette.applications import Starlette


def set_app_attribute(app: Starlette, key: str, value: Any) -> None:
    setattr(app.state, key, value)


def get_app_attribute(app: Any, key: str, default_value: Any = MISSING) -> Any:
    value: Any = getattr(app.state, key, default_value)
    if value is MISSING:
        raise KeyError(f"{key} not found")
    return value



from enum import Enum

__all__ = ["PaitStatus"]


class PaitStatus(Enum):
    """Interface life cycle"""

    undefined = "undefined"
    # APIs are under development and change frequently
    design = "design"
    dev = "dev"

    # The API has been developed, but a few changes will still be made
    integration = "integration"
    complete = "complete"
    test = "test"

    # The API is ready and fully tested and ready to be release
    pre_release = "pre_release"
    release = "release"

    # API exceptions or version iterations in the production environment need to be archived
    abnormal = "abnormal"
    maintenance = "maintenance"
    archive = "archive"
    abandoned = "abandoned"

    def is_deprecated(self) -> bool:
        return self in (
            PaitStatus.abnormal,
            PaitStatus.maintenance,
            PaitStatus.archive,
            PaitStatus.abandoned,
        )



from typing import Callable, List, Optional, Type

from pydantic import BaseModel

from pait.model import PaitCoreModel
from pait.plugin.unified_response import UnifiedResponsePlugin


class SimpleRoute(BaseModel):
    methods: List[str]
    route: Callable
    url: str


def add_route_plugin(simple_route: SimpleRoute, plugin: Type[UnifiedResponsePlugin]) -> None:
    pait_core_model: Optional["PaitCoreModel"] = getattr(simple_route.route, "pait_core_model", None)
    if not pait_core_model:
        raise RuntimeError(f"{simple_route.route} must be a routing function decorated with pait")
    pait_core_model.add_plugin([plugin.build()], [])



import sys
from typing import IO, Any

from flask import Response, send_from_directory

from pait.app.flask.adapter.response import gen_response, set_info_to_response
from pait.plugin.mock_response import MockPluginProtocol

__all__ = ["MockPlugin"]


class MockPlugin(MockPluginProtocol[Response]):
    def get_response(self) -> Response:
        return gen_response(
            self.pait_response_model.get_example_value(example_column_name=self.example_column_name),
            self.pait_response_model,
        )

    def get_file_response(self, temporary_file: IO[bytes], f: Any) -> Response:
        try:
            f.write(self.pait_response_model.get_example_value(example_column_name=self.example_column_name))
            f.seek(0)
            _, f_path, f_filename = temporary_file.name.split("/")
            return send_from_directory("/" + f_path, f_filename, mimetype=self.pait_response_model.media_type)
        finally:
            exc_type, exc_val, exc_tb = sys.exc_info()
            if exc_type is None:
                temporary_file.__exit__(None, None, None)

    def set_info_to_response(self, resp: Response) -> None:
        set_info_to_response(resp, self.pait_response_model)



from typing import Any, Callable, Dict, Generator, List, Optional, Type

from any_api.openapi.model.openapi import Oauth2SecurityModel, OAuthFlowModel, OAuthFlowsModel
from any_api.openapi.model.openapi.security import UserScopesOauth2SecurityModel
from pydantic import BaseModel, Field

from pait.field import Form, Header
from pait.model import JsonResponseModel, PaitCoreModel, get_core_model

from .base import BaseSecurity, SecurityModelType
from .util import get_authorization_scheme_param, set_and_check_field

__all__ = [
    "BaseOAuth2PasswordBearer",
    "BaseOAuth2PasswordRequestFrom",
    "OAuth2PasswordRequestFrom",
    "OAuth2PasswordRequestFromStrict",
    "OAuth2PasswordBearerJsonRespModel",
]


class ScopeType(List[str]):
    @classmethod
    def __get_validators__(cls) -> Generator[Callable, None, None]:
        yield cls.validate

    @classmethod
    def validate(cls, v: str) -> List[str]:
        return v.split(" ")


class BaseOAuth2PasswordRequestFrom(BaseModel):
    username: str = Form()
    password: str = Form()
    scope: ScopeType = Form("")
    client_id: Optional[str] = Form(None)
    client_secret: Optional[str] = Form(None)


class OAuth2PasswordRequestFrom(BaseOAuth2PasswordRequestFrom):
    grant_type: Optional[str] = Form(None, regex="password")


class OAuth2PasswordRequestFromStrict(BaseOAuth2PasswordRequestFrom):
    grant_type: str = Form(regex="password")


class OAuth2PasswordBearerJsonRespModel(JsonResponseModel):
    class ResponseModel(BaseModel):
        access_token: str = Field(description="")
        token_type: str = Field(default="bearer")

    response_data: Type[BaseModel] = ResponseModel


class BaseOAuth2PasswordBearerProxy(BaseSecurity):
    def __init__(self, *, security: "BaseOAuth2PasswordBearer", use_scopes: Optional[List[str]] = None):
        self.security = security
        self.use_scopes = use_scopes
        self.security_name = security.security_name

        def pait_handler(authorization: str = security.header_field) -> str:
            return self.authorization_handler(authorization)

        self.set_pait_handler(pait_handler)

    def __call__(self, authorization: str = Header.i()) -> str:
        raise RuntimeError("should not call this method")  # pragma: no cover

    @property
    def model(self) -> SecurityModelType:
        if self.use_scopes:
            return UserScopesOauth2SecurityModel(
                use_scopes=self.use_scopes, model=self.security.model, flows=self.security.model.flows
            )
        return self.security.model

    def is_allow(self, scopes: List[str]) -> bool:
        return len(set(scopes) & set(self.model.get_security_scope())) > 0

    def authorization_handler(self, authorization: str) -> str:
        scheme, param = get_authorization_scheme_param(authorization)
        if not authorization or scheme.lower() != "bearer":
            raise self.security.not_authenticated_exc
        return param


class BaseOAuth2PasswordBearer(BaseSecurity):
    _proxy = BaseOAuth2PasswordBearerProxy

    def __init__(
        self,
        *,
        route: Optional[Callable] = None,
        security_name: Optional[str] = None,
        scopes: Optional[Dict[str, str]] = None,
        header_field: Optional[Header] = None,
    ):
        self._model: Optional[SecurityModelType] = None
        self._scopes: Dict[str, str] = scopes or {}

        self.route: Optional[Callable] = None
        self.security_name = security_name or self.__class__.__name__
        self.header_field: Header = header_field or Header.i(openapi_include=False)

        if route:
            self.with_route(route)

        # init field
        self.not_authenticated_exc: Exception = self.get_exception(
            status_code=401, message="Not authenticated", headers={"WWW-Authenticate": "Bearer"}
        )
        set_and_check_field(self.header_field, "Authorization", self.not_authenticated_exc)

    def with_route(self, route: Callable) -> "BaseOAuth2PasswordBearer":
        if self.route is not None:
            raise ValueError("route has been set")
        self.route = route
        core_model: PaitCoreModel = get_core_model(route)
        self._set_model(core_model.path)

        def _set_model_by_core_model_change(*args: Any) -> None:
            self._set_model(core_model.path)

        core_model.add_change_notify(_set_model_by_core_model_change)
        return self

    def _set_model(self, path: str) -> None:
        self._model = Oauth2SecurityModel(
            flows=OAuthFlowsModel(password=OAuthFlowModel(tokenUrl=path, scopes=self._scopes))
        )

    @property
    def model(self) -> SecurityModelType:
        if self._model is None:
            raise ValueError("The model is invalid, please use the `with_route` method to set the routing function")
        return self._model

    def get_depend(self, use_scopes: Optional[List[str]] = None) -> "BaseOAuth2PasswordBearerProxy":
        self.model  # It is only used to check whether a route is bound
        return self._proxy(
            security=self,
            use_scopes=use_scopes,
        )



import asyncio
from contextvars import copy_context
from functools import partial
from typing import Any

from pait.app.tornado.adapter.response import gen_response
from pait.plugin.base import PluginContext
from pait.plugin.unified_response import UnifiedResponsePlugin as BaseUnifiedResponsePlugin
from pait.plugin.unified_response import UnifiedResponsePluginProtocol as BaseUnifiedResponsePluginProtocol


def _gen_response(self: BaseUnifiedResponsePluginProtocol, return_value: Any, context: PluginContext) -> Any:
    return gen_response(return_value, self.response_model_class, context.cbv_instance, *context.args, **context.kwargs)


class UnifiedResponsePlugin(BaseUnifiedResponsePlugin):
    _gen_response = _gen_response

    async def __call__(self, context: PluginContext) -> Any:
        # Compatible with tornado cannot call sync routing function
        if self._is_async_func:
            return await self._async_call(context)
        else:
            return await asyncio.get_event_loop().run_in_executor(
                None, partial(copy_context().run, partial(self._sync_call, context))  # Matryoshka
            )


class UnifiedResponsePluginProtocol(BaseUnifiedResponsePluginProtocol):
    _gen_response = _gen_response



from dataclasses import MISSING
from typing import Any, List, Mapping

from tornado.httputil import HTTPServerRequest
from tornado.web import RequestHandler

from pait.app.base import BaseAppHelper
from pait.app.tornado._attribute import get_app_attribute
from pait.app.tornado.adapter.request import Request, RequestExtend

__all__ = ["AppHelper", "RequestExtend"]


class AppHelper(BaseAppHelper[HTTPServerRequest, RequestExtend]):
    CbvType = (RequestHandler,)
    app_name = "tornado"

    request_class = Request

    def __init__(self, args: List[Any], kwargs: Mapping[str, Any]):
        super().__init__(args, kwargs)
        if not self.cbv_instance:
            raise RuntimeError("Can not load Tornado handle")  # pragma: no cover

    def _get_request(self, args: List[Any]) -> HTTPServerRequest:
        return self.cbv_instance.request

    def get_attributes(self, key: str, default: Any = MISSING) -> Any:
        return get_app_attribute(self.cbv_instance.application, key, default)



from dataclasses import MISSING
from typing import Any

from tornado.web import Application


def set_app_attribute(app: Application, key: str, value: Any) -> None:
    app.settings[key] = value


def get_app_attribute(app: Application, key: str, default_value: Any = MISSING) -> Any:
    value: Any = app.settings.get(key, default_value)
    if value is MISSING:
        raise KeyError(f"{key} not found")
    return value



import re
from typing import TYPE_CHECKING, Any, Callable, List, Optional, Set, Tuple, Type

from pydantic import BaseConfig, BaseModel

from pait.model.response import BaseResponseModel
from pait.plugin.base import PluginManager, PrePluginProtocol
from pait.types import Literal
from pait.util import http_method_tuple

if TYPE_CHECKING:
    from pait.model import APPLY_FN, PaitCoreModel
    from pait.param_handle import BaseParamHandler


__all__ = [
    "apply_multi_plugin",
    "apply_extra_openapi_model",
    "apply_response_model",
    "apply_default_pydantic_model_config",
    "apply_block_http_method_set",
    "apply_pre_depend",
    "MatchRule",
]


MatchKeyLiteral = Literal[
    "all",
    "status",
    "group",
    "tag",
    "method_list",
    "path",
    "!status",
    "!group",
    "!tag",
    "!method_list",
    "!path",
]


class MatchRule(object):
    def __init__(self, key: MatchKeyLiteral = "all", target: Any = None):
        self.key: MatchKeyLiteral = key
        self.target: Any = target
        self._match_rule_list: List[Tuple[str, "MatchRule"]] = []

    def match(self, pait_core_model: "PaitCoreModel") -> bool:
        result: bool = self._match(pait_core_model)
        for method, match_rule in self._match_rule_list:
            if method == "or":
                result = result or match_rule.match(pait_core_model)
            elif method == "and":
                result = result and match_rule.match(pait_core_model)
        return result

    def _match(self, pait_core_model: "PaitCoreModel") -> bool:
        """By different attributes to determine whether to match with pait_core_model,
        if the key is `all` then match
        if the key is prefixed with ! then the result will be reversed
        """
        key: MatchKeyLiteral = self.key
        target: Any = self.target
        if key == "all":
            return True
        is_reverse: bool = False
        if key.startswith("!"):
            key = key[1:]  # type: ignore
            is_reverse = True

        value: Any = getattr(pait_core_model, key, ...)
        if value is ...:
            raise KeyError(f"match fail, not found key:{key}")
        if key in ("status", "group"):
            result: bool = value is target
        elif key in ("tag", "method_list"):
            result = target in value
        elif key == "path":
            result = bool(re.match(target, value))
        else:
            raise KeyError(f"Not support key:{key}")

        if is_reverse:
            return not result
        else:
            return result

    def __or__(self, other: "MatchRule") -> "MatchRule":
        self._match_rule_list.append(("or", other))
        return self

    def __and__(self, other: "MatchRule") -> "MatchRule":
        self._match_rule_list.append(("and", other))
        return self

    def __repr__(self) -> str:
        return f"<MatchRule object; key={self.key}, target={self.target} match_rule_list={self._match_rule_list}>"


def _is_match(pait_core_model: "PaitCoreModel", match_rule: Optional["MatchRule"]) -> bool:
    return not match_rule or match_rule.match(pait_core_model)


def apply_extra_openapi_model(
    extra_openapi_model: Type[BaseModel], match_rule: Optional["MatchRule"] = None
) -> "APPLY_FN":
    """
    Add a default extre_openapi structure for routing handles
    """

    def _apply(pait_core_model: "PaitCoreModel") -> None:
        if _is_match(pait_core_model, match_rule):
            pait_core_model.extra_openapi_model_list = [extra_openapi_model]

    return _apply


def apply_response_model(
    response_model_list: List[Type[BaseResponseModel]], match_rule: Optional["MatchRule"] = None
) -> "APPLY_FN":
    """
    Add a default response structure for routing handles
    """

    def _apply(pait_core_model: "PaitCoreModel") -> None:
        if _is_match(pait_core_model, match_rule):
            pait_core_model.add_response_model_list(response_model_list)

    return _apply


def apply_default_pydantic_model_config(
    pydantic_model_config: Type[BaseConfig], match_rule: Optional["MatchRule"] = None
) -> "APPLY_FN":
    """pait route gen pydantic model default config"""

    def _apply(pait_core_model: "PaitCoreModel") -> None:
        if _is_match(pait_core_model, match_rule):
            pait_core_model.pydantic_model_config = pydantic_model_config

    return _apply


def apply_default_pydantic_basemodel(
    pydantic_basemodel: Type[BaseModel], match_rule: Optional["MatchRule"] = None
) -> "APPLY_FN":
    """pait route gen pydantic model default pydantic base model"""

    def _apply(pait_core_model: "PaitCoreModel") -> None:
        if _is_match(pait_core_model, match_rule):
            pait_core_model.pydantic_basemodel = pydantic_basemodel

    return _apply


def apply_block_http_method_set(
    block_http_method_set: Set[str], match_rule: Optional["MatchRule"] = None
) -> "APPLY_FN":
    """
    Under normal circumstances, pait.load_app can obtain the http method of the routing handle.
     However, some application frameworks such as flask will automatically add optional http methods
     to the handle, which is great, but you may not want to use them in pait, and pait will not
     automatically recognize them, so you can use parameters to disable certain http method
    """
    for block_http_method in block_http_method_set:
        if block_http_method.lower() not in http_method_tuple:
            raise ValueError(f"Error http method: {block_http_method}")

    def _apply(pait_core_model: "PaitCoreModel") -> None:
        if _is_match(pait_core_model, match_rule):
            pait_core_model.block_http_method_set = block_http_method_set
            pait_core_model.method_list = pait_core_model.method_list

    return _apply


def apply_multi_plugin(
    plugin_manager_fn_list: List[Callable[[], PluginManager]], match_rule: Optional["MatchRule"] = None
) -> "APPLY_FN":
    def _apply(pait_core_model: "PaitCoreModel") -> None:
        if _is_match(pait_core_model, match_rule):
            pre_plugin_manager_list: List[PluginManager] = []
            post_plugin_manager_list: List[PluginManager] = []
            for plugin_manager_fn in plugin_manager_fn_list:
                plugin_manager: PluginManager = plugin_manager_fn()
                if issubclass(plugin_manager.plugin_class, PrePluginProtocol):
                    pre_plugin_manager_list.append(plugin_manager)
                else:
                    post_plugin_manager_list.append(plugin_manager)
            pait_core_model.add_plugin(pre_plugin_manager_list, post_plugin_manager_list)

    return _apply


def apply_pre_depend(pre_depend: Callable, match_rule: Optional["MatchRule"] = None) -> "APPLY_FN":
    def _apply(pait_core_model: "PaitCoreModel") -> None:
        if _is_match(pait_core_model, match_rule):
            pait_core_model.pre_depend_list.append(pre_depend)

    return _apply


def apply_param_handler(
    param_handler_plugin: "Type[BaseParamHandler]", match_rule: Optional["MatchRule"] = None
) -> "APPLY_FN":
    def _apply(pait_core_model: "PaitCoreModel") -> None:
        if _is_match(pait_core_model, match_rule):
            pait_core_model.param_handler_plugin = param_handler_plugin

    return _apply



from importlib import import_module
from typing import Any, Callable, List

from typing_extensions import Literal

SupportAppLiteral = Literal["flask", "starlette", "sanic", "tornado"]
support_app_list: List[SupportAppLiteral] = ["flask", "starlette", "sanic", "tornado"]


def sniffing(app: Any) -> SupportAppLiteral:
    app_name: str = app.__class__.__name__.lower()
    if app_name in support_app_list:
        return app_name  # type: ignore
    elif app_name == "application" and app.__class__.__module__ == "tornado.web":
        return "tornado"
    else:
        raise NotImplementedError(f"Pait not support app name:{app_name}, please check app:{app}")


def import_func_from_app(fun_name: str, app: Any = None, module_name: str = "") -> Callable:
    from pait.app.auto_load_app import auto_load_app_class

    if app:
        # support werkzeug LocalProxy
        _get_current_object = getattr(app, "_get_current_object", None)
        if _get_current_object:
            app = _get_current_object()
        app_name: str = sniffing(app)
    else:
        app_name = auto_load_app_class().__name__.lower()
    if module_name:
        return getattr(import_module(f"pait.app.{app_name}.{module_name}"), fun_name)
    else:
        return getattr(import_module(f"pait.app.{app_name}"), fun_name)


def base_call_func(fun_name: str, *args: Any, app: Any = None, module_name: str = "", **kwargs: Any) -> Any:
    return import_func_from_app(fun_name, app=app, module_name=module_name)(*args, **kwargs)



from typing import Any, Optional

from werkzeug.exceptions import HTTPException
from werkzeug.wrappers.response import Response


class HTTPExceptionWithHeaders(HTTPException):
    headers: Optional[dict] = None

    def set_headers(self, headers: Optional[dict] = None) -> None:
        self.headers = headers

    def __call__(self, environ: Any, start_response: Any) -> Response:
        response: Response = self.get_response(environ)
        if self.headers:
            response.headers.update(self.headers)
        return response(environ, start_response)


def http_exception(*, status_code: int, message: str, headers: Optional[dict] = None) -> Exception:
    exc = HTTPExceptionWithHeaders(description=message)
    exc.set_headers(headers)
    exc.code = status_code
    return exc



import inspect
import logging
from typing import TYPE_CHECKING, Any, Dict, List, Mapping, Optional, Type, Union

from pydantic import BaseConfig, BaseModel
from pydantic.error_wrappers import ValidationError
from pydantic.fields import ModelField, Undefined, UndefinedType
from pydantic.schema import get_annotation_from_field_info
from typing_extensions import Self  # type: ignore

from pait import field
from pait.exceptions import (
    FieldValueTypeException,
    NotFoundFieldException,
    NotFoundValueException,
    PaitBaseException,
    ParseTypeError,
    TipException,
)
from pait.field import BaseField
from pait.plugin.base import PluginProtocol
from pait.types import CallType
from pait.util import (
    FuncSig,
    create_pydantic_model,
    example_value_handle,
    gen_tip_exc,
    get_func_sig,
    get_parameter_list_from_pydantic_basemodel,
    ignore_pre_check,
    is_bounded_func,
    is_type,
)

if TYPE_CHECKING:
    from pait.model.context import ContextModel
    from pait.model.core import PaitCoreModel


def raise_multiple_exc(exc_list: List[Exception]) -> None:
    """
    Multiple exceptions may be thrown during the parsing process, and these will be thrown one by one like a stack
    """
    if not exc_list:
        return
    try:
        raise exc_list.pop()
    finally:
        raise_multiple_exc(exc_list)


class BaseParamHandler(PluginProtocol):
    tip_exception_class: Optional[Type[TipException]] = TipException

    # def __post_init__(self, pait_core_model: "PaitCoreModel", args: tuple, kwargs: dict) -> None:
    #     super(BaseParamHandler, self).__post_init__(pait_core_model, args, kwargs)

    #     # cbv handle
    #     context_model: "ContextModel" = pait_context.get()
    #     self.cbv_instance: Optional[Any] = context_model.cbv_instance
    #     self._app_helper: "BaseAppHelper" = context_model.app_helper
    #     self.cbv_type: Optional[Type] = None

    #     if self.cbv_instance:
    #         self.cbv_type = self.cbv_instance.__class__
    #         self.cbv_param_list: List["inspect.Parameter"] = get_parameter_list_from_class(self.cbv_type)
    #     else:
    #         self.cbv_param_list = []
    #         # cbv_type = getattr(
    #               inspect.getmodule(func),
    #               func.__qualname__.split(".<locals>", 1)[0].rsplit(".", 1)[0]
    #               )
    #     self.pre_depend_list: List[Callable] = pait_core_model.pre_depend_list
    #     self.pydantic_model_config: Type[BaseConfig] = pait_core_model.pydantic_model_config

    @classmethod
    def check_depend_handle(cls, pait_core_model: "PaitCoreModel", func: CallType) -> Any:
        if inspect.ismethod(func) and not is_bounded_func(func):
            raise ValueError(f"Method: {func.__qualname__} is not a bounded function")  # pragma: no cover
        func_sig: FuncSig = get_func_sig(func)  # get and cache func sig
        cls.check_param_field_handle(pait_core_model, func_sig, func_sig.param_list)

    @staticmethod
    def check_field_type(value: Any, target_type: Any, error_msg: str) -> None:
        if value is ...:
            return
        if isinstance(value, UndefinedType):
            return
        if inspect.isclass(target_type) and isinstance(value, target_type):
            return
        if getattr(value, "__call__", None):
            value = value()
        try:
            create_pydantic_model({"faker_param_name": (target_type, ...)})(faker_param_name=value)
        except Exception:
            raise ParseTypeError(error_msg)

    @classmethod
    def check_param_field_by_parameter(
        cls,
        pait_core_model: "PaitCoreModel",
        parameter: inspect.Parameter,
    ) -> None:
        if isinstance(parameter.default, field.Depends):
            cls.check_depend_handle(pait_core_model, parameter.default.func)
            if ignore_pre_check:
                return
            func_sig: FuncSig = get_func_sig(parameter.default.func)  # get and cache func sig
            if not is_type(parameter.annotation, func_sig.return_param):
                raise FieldValueTypeException(
                    parameter.name,
                    f"{parameter.name}'s Depends.callable return annotation"
                    f" must:{parameter.annotation}, not {func_sig.return_param}",
                )
        elif isinstance(parameter.default, field.BaseField):
            if not parameter.default.alias:
                parameter.default.request_key = parameter.name
            if ignore_pre_check:
                return
            if parameter.default.alias and not isinstance(parameter.default.alias, str):
                raise FieldValueTypeException(
                    parameter.name,
                    f"{parameter.name}'s Field.alias type must str. value: {parameter.default.alias}",
                )
            if parameter.default.get_field_name() not in pait_core_model.app_helper_class.request_class.__dict__:
                raise NotFoundFieldException(
                    parameter.name,
                    f"field name: {parameter.default.get_field_name()}"
                    f" not found in {pait_core_model.app_helper_class.request_class}",
                )  # pragma: no cover
            try:
                cls.check_field_type(
                    parameter.default.default,
                    parameter.annotation,
                    f"{parameter.name}'s Field.default type must {parameter.annotation}."
                    f" value:{parameter.default.default}",
                )
                if parameter.default.default_factory:
                    cls.check_field_type(
                        parameter.default.default_factory(),
                        parameter.annotation,
                        f"{parameter.name}'s Field.default_factory type must {parameter.annotation}."
                        f" value:{parameter.default.default_factory()}",
                    )
                example_value: Any = parameter.default.extra.get("example", Undefined)
                cls.check_field_type(
                    example_value_handle(example_value),
                    parameter.annotation,
                    f"{parameter.name}'s Field.example type must {parameter.annotation} not {example_value}",
                )
            except ParseTypeError as e:
                raise FieldValueTypeException(parameter.name, str(e))
        else:
            raise NotFoundFieldException(parameter.name, f"{parameter.name}'s Field not found")  # pragma: no cover

    @classmethod
    def check_param_field_handle(
        cls,
        pait_core_model: "PaitCoreModel",
        _object: Union[FuncSig, Type, None],
        param_list: List["inspect.Parameter"],
    ) -> None:
        for parameter in param_list:
            try:
                if parameter.default != parameter.empty:
                    # kwargs param
                    # support model: def demo(pydantic.BaseModel: BaseModel = pait.field.BaseField())
                    cls.check_param_field_by_parameter(pait_core_model, parameter)
                else:
                    # args param
                    # support model: model: ModelType
                    if parameter.annotation == Self:
                        continue
                    if not (inspect.isclass(parameter.annotation) and issubclass(parameter.annotation, BaseModel)):
                        continue
                    # cache and get parameter_list
                    param_list = get_parameter_list_from_pydantic_basemodel(
                        parameter.annotation, default_field_class=pait_core_model.default_field_class
                    )
                    for _parameter in param_list:
                        cls.check_param_field_by_parameter(pait_core_model, _parameter)
            except PaitBaseException as e:
                raise gen_tip_exc(_object, e, parameter, tip_exception_class=cls.tip_exception_class) from e

    @classmethod
    def pre_hook(cls, pait_core_model: "PaitCoreModel", kwargs: Dict) -> None:
        # check and load param from pre depend
        for pre_depend in pait_core_model.pre_depend_list:
            cls.check_depend_handle(pait_core_model, pre_depend)

        # check and load param from func
        func_sig: FuncSig = get_func_sig(pait_core_model.func)
        cls.check_param_field_handle(pait_core_model, func_sig, func_sig.param_list)

        # TODO support cbv class Attribute
        # I don't know how to get the class of the decorated function at the initialization of the decorator,
        # which may be an unattainable requirement

    @classmethod
    def pre_load_hook(cls, pait_core_model: "PaitCoreModel", kwargs: Dict) -> Dict:
        super().pre_load_hook(pait_core_model, kwargs)
        if ignore_pre_check:
            # pre_check has helped to do the same task as pre_load
            cls.pre_hook(pait_core_model, kwargs)  # pragma: no cover
        return kwargs

    @classmethod
    def pre_check_hook(cls, pait_core_model: "PaitCoreModel", kwargs: Dict) -> None:
        super().pre_check_hook(pait_core_model, kwargs)
        cls.pre_hook(pait_core_model, kwargs)

    def _set_parameter_value_to_args(
        self, context: "ContextModel", parameter: inspect.Parameter, func_args: list
    ) -> bool:
        """Extract the self parameter of the cbv handler,
        the request parameter of the route and the parameter of type PaitBaseModel,
        and check if there are any other parameters that do not meet the conditions

        Sort by frequency of occurrence
        """
        if context.cbv_instance and (
            parameter.annotation == Self or (not func_args and parameter.annotation == parameter.empty)
        ):
            # first parma must self param, looking forward to the appearance of `self type
            func_args.append(context.cbv_instance)
        elif inspect.isclass(parameter.annotation) and issubclass(parameter.annotation, BaseModel):
            return True
        elif context.app_helper.request.check_request_type(parameter.annotation):
            # support request param(def handle(request: Request))
            func_args.append(context.app_helper.request.request)
        else:
            logging.warning(f"Pait not support args: {parameter}")  # pragma: no cover
        return False

    @staticmethod
    def request_value_handle(
        parameter: inspect.Parameter,
        request_value: Mapping,
        kwargs_param_dict: Dict[str, Any],
        pydantic_model: Optional[Type[BaseModel]] = None,
    ) -> None:
        """parse request_value and set to base_model_dict or parameter_value_dict"""
        pait_field: BaseField = parameter.default
        annotation: Type[BaseModel] = parameter.annotation

        if not pait_field.raw_return:
            request_value = pait_field.request_value_handle(request_value)
            if request_value is Undefined:
                raise (
                    pait_field.not_value_exception
                    or NotFoundValueException(parameter.name, f"Can not found {parameter.name} value")
                )
        if pydantic_model:
            # If it belongs to the field of the PaitModel,
            # it does not need to be verified, but is returned to the PaitModel for processing
            kwargs_param_dict[pait_field.request_key] = request_value
        elif inspect.isclass(annotation) and issubclass(annotation, BaseModel):
            # parse annotation is pydantic.BaseModel and base_model_dict not None
            kwargs_param_dict[parameter.name] = annotation(**request_value)
        else:
            # parse annotation is python type and pydantic.field
            request_value, e = ModelField(
                name=parameter.name,
                type_=get_annotation_from_field_info(annotation, pait_field, parameter.name),
                model_config=BaseConfig,
                field_info=pait_field,
                class_validators={},
                # Since the corresponding value has already been processed, there is no need to process it again
                # default: Any = None,
                # default_factory: Optional[NoArgAnyCallable] = None,
                # required: 'BoolUndefined' = Undefined,
                # final: bool = False,
                # alias: Optional[str] = None,
            ).validate(request_value, kwargs_param_dict, loc=(pait_field.get_field_name(), parameter.name))
            if e:
                raise ValidationError([e], BaseModel)
            kwargs_param_dict[parameter.name] = request_value



from typing import Any, Dict, List, Mapping

from flask import Request as FlaskRequest
from flask import request as _request
from werkzeug.datastructures import EnvironHeaders, ImmutableMultiDict

from pait.app.base.adapter.request import BaseRequest, BaseRequestExtend
from pait.util import LazyProperty


class RequestExtend(BaseRequestExtend[FlaskRequest]):
    @property
    def scheme(self) -> str:
        return _request.scheme

    @property
    def path(self) -> str:
        return _request.path

    @property
    def hostname(self) -> str:
        return _request.host


class Request(BaseRequest[FlaskRequest, RequestExtend]):
    RequestType = FlaskRequest
    FormType = ImmutableMultiDict
    FileType = FlaskRequest.files
    HeaderType = EnvironHeaders

    def request_extend(self) -> RequestExtend:
        return RequestExtend(self.request)

    def body(self) -> dict:
        return _request.json

    def cookie(self) -> dict:
        return _request.cookies

    def file(self) -> FlaskRequest.files:  # type: ignore
        return _request.files

    def form(self) -> FlaskRequest.form:  # type: ignore
        return _request.form

    def header(self) -> EnvironHeaders:
        return _request.headers

    def path(self) -> Mapping[str, Any]:
        return self.request_kwargs

    def query(self) -> Dict[str, Any]:
        return _request.args

    @LazyProperty()
    def multiform(self) -> Dict[str, List[Any]]:
        return {key: _request.form.getlist(key) for key, _ in _request.form.items()}

    @LazyProperty()
    def multiquery(self) -> Dict[str, List[Any]]:
        return {key: _request.args.getlist(key) for key, _ in _request.args.items()}



from typing import Optional

from starlette.exceptions import HTTPException


def http_exception(*, status_code: int, message: str, headers: Optional[dict] = None) -> Exception:
    exc: HTTPException = HTTPException(detail=message, status_code=status_code)
    if headers:
        exc.headers = headers
    return exc



from typing import Any

import aiofiles  # type: ignore
from tornado.web import RequestHandler

from pait.app.tornado.adapter.response import gen_response, set_info_to_response
from pait.plugin.base import PluginContext
from pait.plugin.mock_response import MockPluginProtocol


class MockPlugin(MockPluginProtocol[None]):
    tornado_handle: RequestHandler

    def get_response(self) -> None:
        return gen_response(
            self.pait_response_model.get_example_value(example_column_name=self.example_column_name),
            self.pait_response_model,
            self.tornado_handle,
        )

    def set_info_to_response(self, resp: None) -> None:
        set_info_to_response(self.tornado_handle, self.pait_response_model)

    async def async_get_file_response(self, temporary_file: Any, f: Any) -> None:
        await f.write(self.pait_response_model.get_example_value(example_column_name=self.example_column_name))
        await f.seek(0)
        async for line in f:
            self.tornado_handle.write(line)
        await temporary_file.__aexit__(None, None, None)

    async def __call__(self, context: PluginContext) -> Any:
        self.tornado_handle = context.args[0]
        await super().__call__(context)



from dataclasses import MISSING
from typing import Any

from starlette.endpoints import HTTPEndpoint
from starlette.requests import Request as _Request

from pait.app.base import BaseAppHelper
from pait.app.starlette._attribute import get_app_attribute
from pait.app.starlette.adapter.request import Request, RequestExtend

__all__ = ["AppHelper", "RequestExtend"]


class AppHelper(BaseAppHelper[_Request, RequestExtend]):
    CbvType = (HTTPEndpoint,)
    app_name = "starlette"

    request_class = Request

    def get_attributes(self, key: str, default: Any = MISSING) -> Any:
        return get_app_attribute(self.raw_request.app, key, default)



from typing import Dict, Optional

from pait.app.tornado import http_exception


class GetException(object):
    @classmethod
    def get_exception(cls, *, status_code: int, message: str, headers: Optional[Dict] = None) -> Exception:
        return http_exception(status_code=status_code, message=message, headers=headers)



from typing import Any

from pait.plugin.auto_complete_json_resp import AutoCompleteJsonRespPlugin as _AutoCompleteJsonRespPlugin
from pait.plugin.base import PluginContext

from .unified_response import UnifiedResponsePluginProtocol

__all__ = ["AutoCompleteJsonRespPlugin"]


class AutoCompleteJsonRespPlugin(UnifiedResponsePluginProtocol, _AutoCompleteJsonRespPlugin):
    def _sync_call(self, context: PluginContext) -> Any:
        response: Any = super()._sync_call(context)
        return self._gen_response(response, context)

    async def _async_call(self, context: PluginContext) -> Any:
        response: Any = await super()._async_call(context)
        return self._gen_response(response, context)



import inspect
import typing
from typing import Any, Generator, List, Optional, Set, Type, Union, _GenericAlias  # type: ignore

import typing_extensions

from pait.exceptions import ParseTypeError

_PYTHON_ORIGIN_TYPE_SET: Set[Optional[type]] = {bool, dict, float, int, list, str, tuple, type(None), None, set}
_TYPING_NOT_PARSE_TYPE_SET: set = {
    typing.Any,
    typing.Generic,
    typing_extensions.Never,
    typing_extensions.NoReturn,
    typing_extensions.Self,
    typing_extensions.ClassVar,
    typing_extensions.Final,
}
__all__ = ["parse_typing", "is_type"]


def parse_typing(_type: Any) -> List[Type]:
    """Get Python Type through typing as much as possible
    >>> assert [dict] == parse_typing(dict)
    >>> assert [list] == parse_typing(List)
    >>> assert [list] == parse_typing(List[str])
    >>> assert [dict] == parse_typing(Dict)
    >>> assert [set] == parse_typing(typing.Set)
    >>> assert [dict] == parse_typing(Union[dict])
    >>> assert [dict] == parse_typing(Union[Dict])
    >>> assert [dict] == parse_typing(Union[Dict[str, Any]])
    >>> assert [dict] == parse_typing(typing.AsyncIterator[Dict])
    >>> assert [dict] == parse_typing(typing.Iterator[Dict])
    >>> assert [dict] == parse_typing(typing.Generator[Dict, None, None])
    >>> assert [dict] == parse_typing(typing.AsyncGenerator[Dict, None])
    >>> assert [tuple] == parse_typing(typing.Tuple[str, int])
    >>> assert [dict] == parse_typing(typing_extensions.TypedDict)
    >>> # multi value
    >>> assert [dict, str, int] == parse_typing(Union[Dict, str, int])
    >>> assert [dict, type(None)] == parse_typing(Optional[Dict])
    >>> assert [dict, type(None)] == parse_typing(Optional[dict])
    >>> # Relatively rarely used scenes
    >>> assert [int] == parse_typing(typing.NewType("UserId", int))
    >>> assert [int] == parse_typing(typing.Callable[[], int])
    >>> assert [int] == parse_typing(typing.Callable[[], typing.Awaitable[int]])
    >>> s = typing.TypeVar("s", int, str)
    >>> assert [int, str] == parse_typing(s)
    >>> assert [str] == parse_typing(typing_extensions.Literal["a", "b"])
    >>> assert [str] == parse_typing(typing_extensions.LiteralString)

    >>> class Demo:
    >>>     pass

    >>> assert [int] == parse_typing(typing_extensions.Annotated[int, Demo])
    >>> # Unresolved type, returned directly
    >>> from pait.util._types import _TYPING_NOT_PARSE_TYPE_SET

    >>> for i in _TYPING_NOT_PARSE_TYPE_SET:
    >>>     assert [i] == parse_typing(i)
    """
    origin: Optional[type] = getattr(_type, "__origin__", None)  # get typing.xxx's raw type
    if origin:
        if origin is Union:
            # support Union, Optional
            type_list: List[Type[Any]] = []
            for i in _type.__args__:
                type_list.extend(parse_typing((i)))
            return type_list
        elif origin is typing_extensions.Literal:
            return [str]
        elif origin in _PYTHON_ORIGIN_TYPE_SET:
            return [origin]
        arg_list: List = getattr(_type, "__args__", [])
        if arg_list:
            # support AsyncIterator, Iterator
            return parse_typing(arg_list[0])
        return [origin]
    elif _type is Generator:
        return parse_typing(_type.__args__[0])
    elif _type in _PYTHON_ORIGIN_TYPE_SET:
        return [_type]
    elif getattr(_type, "_name", "") == "Optional":
        return [type(None)]
    elif getattr(_type, "_name", "") == "LiteralString":
        return [str]
    elif typing_extensions.is_typeddict(_type):
        return [dict]
    elif hasattr(_type, "__supertype__"):
        # support NewType
        return [getattr(_type, "__supertype__")]
    elif hasattr(_type, "__constraints__"):
        # support TypeVar
        return list(getattr(_type, "__constraints__"))
    elif inspect.isclass(_type):
        return [_type]
    elif _type in _TYPING_NOT_PARSE_TYPE_SET:
        return [_type]
    else:
        raise ParseTypeError(f"Can not parse {_type} origin type")


def is_type(source_type: Type, target_type: Union[Type, object]) -> bool:
    """Determine whether the two types are consistent"""
    return bool(set(parse_typing(target_type)) & set(parse_typing(source_type)))



import copy
import logging
from typing import TYPE_CHECKING, Any, Callable, List, Optional, Set, Tuple, Type

from pydantic import BaseConfig, BaseModel

from pait.model.response import BaseResponseModel, PaitResponseModel
from pait.model.status import PaitStatus
from pait.model.tag import Tag
from pait.param_handle import BaseParamHandler
from pait.plugin import PluginManager, PluginProtocol, PostPluginProtocol, PrePluginProtocol
from pait.util import gen_tip_exc, ignore_pre_check

if TYPE_CHECKING:
    from pait.app.base import BaseAppHelper
    from pait.field import BaseField


__all__ = ["PaitCoreModel", "get_core_model"]
ChangeNotifyType = Callable[["PaitCoreModel", str, Any], None]


def get_core_model(route: Callable) -> "PaitCoreModel":
    core_model: Optional["PaitCoreModel"] = getattr(route, "pait_core_model", None)
    if not core_model:
        raise TypeError(f"Routing function: {route} has not been wrapped by pait")
    return core_model


class PaitCoreModel(object):
    _param_handler_plugin: PluginManager["BaseParamHandler"]
    _main_plugin: PluginProtocol

    def __init__(
        self,
        func: Callable,
        app_helper_class: "Type[BaseAppHelper]",
        param_handler_plugin: Type[BaseParamHandler],
        pre_depend_list: Optional[List[Callable]] = None,
        path: Optional[str] = None,
        openapi_path: Optional[str] = None,
        method_set: Optional[Set[str]] = None,
        operation_id: Optional[str] = None,
        func_name: Optional[str] = None,
        author: Optional[Tuple[str, ...]] = None,
        summary: Optional[str] = None,
        desc: Optional[str] = None,
        status: Optional[PaitStatus] = None,
        group: Optional[str] = None,
        tag: Optional[Tuple[Tag, ...]] = None,
        response_model_list: Optional[List[Type[BaseResponseModel]]] = None,
        pydantic_model_config: Optional[Type[BaseConfig]] = None,
        pydantic_basemodel: Optional[Type[BaseModel]] = None,
        default_field_class: Optional[Type["BaseField"]] = None,
        plugin_list: Optional[List[PluginManager[PrePluginProtocol]]] = None,
        post_plugin_list: Optional[List[PluginManager[PostPluginProtocol]]] = None,
        feature_code: str = "",
        **kwargs: Any,
    ):
        # pait
        self.app_helper_class: "Type[BaseAppHelper]" = app_helper_class
        self.default_field_class: Optional[Type["BaseField"]] = default_field_class
        self.func: Callable = func  # route func
        self.pait_id: str = f"{func.__module__}_{func.__qualname__}"
        # Some functions have the same md5 as the name and need to be distinguished by the feature code
        if feature_code:
            self.pait_id = f"{feature_code}_{self.pait_id}"
        setattr(func, "_pait_id", self.pait_id)
        setattr(func, "pait_core_model", self)
        self.pre_depend_list: List[Callable] = pre_depend_list or []
        self.func_path: str = ""
        self.block_http_method_set: Set[str] = set()
        # TODO
        self.pydantic_model_config: Optional[Type[BaseConfig]] = pydantic_model_config
        # TODO
        self.pydantic_basemodel: Optional[Type[BaseModel]] = pydantic_basemodel
        self.extra: dict = kwargs

        # api doc
        self.path: str = path or ""  # request url path
        self.openapi_path: str = openapi_path or ""
        self._method_list: List[str] = sorted(list(method_set or set()))  # request method set
        self.func_name: str = func_name or func.__name__
        self.operation_id: str = operation_id or self.func_name  # route name
        self.author: Optional[Tuple[str, ...]] = author  # The main developer of this func
        self.summary: str = summary or ""
        self.desc: str = desc or func.__doc__ or ""  # desc of this func
        self.status: PaitStatus = status or PaitStatus.undefined
        self.group: str = group or "root"  # Which group this interface belongs to
        self.tag: Tuple[Tag, ...] = tag or (Tag(name="default"),)  # Interface tag
        self._extra_openapi_model_list: List[Type[BaseModel]] = []
        self._response_model_list: List[Type[BaseResponseModel]] = []
        if response_model_list:
            self.add_response_model_list(response_model_list)

        # pait plugin
        self._plugin_list: List[PluginManager] = []
        self._post_plugin_list: List[PluginManager] = []
        self.param_handler_plugin = param_handler_plugin  # type: ignore
        self.add_plugin(plugin_list, post_plugin_list)

        # change notify
        self._change_notify_list: List[ChangeNotifyType] = []

    def add_change_notify(self, callback: ChangeNotifyType) -> None:
        self._change_notify_list.append(callback)

    def remove_change_notify(self, callback: ChangeNotifyType) -> None:
        self._change_notify_list.remove(callback)

    def __setattr__(self, key: str, value: Any) -> None:
        if key.startswith("_"):
            return super().__setattr__(key, value)

        change_notify_list: List[ChangeNotifyType] = self.__dict__.get("_change_notify_list", [])
        if change_notify_list:
            for callback in self._change_notify_list:
                callback(self, key, value)
        return super().__setattr__(key, value)

    @property
    def param_handler_plugin(self) -> Type[BaseParamHandler]:
        return self._param_handler_plugin.plugin_class

    @param_handler_plugin.setter
    def param_handler_plugin(self, param_handler_plugin: Type[BaseParamHandler]) -> None:
        self._param_handler_plugin = PluginManager(param_handler_plugin)

        try:
            if not ignore_pre_check:
                self._param_handler_plugin.pre_check_hook(self)
            self._param_handler_plugin.pre_load_hook(self)
        except Exception as e:
            raise gen_tip_exc(self.func, RuntimeError(f"set param plugin error: {e}")) from e
        self.add_plugin([], [])

    @property
    def method_list(self) -> List[str]:
        _temp_set: Set[str] = set(self._method_list.copy())
        _temp_set.difference_update(self.block_http_method_set)
        return sorted(list(_temp_set))

    @method_list.setter
    def method_list(self, method_list: List[str]) -> None:
        self._method_list = list(set(self._method_list) | set(method_list))

    @property
    def openapi_method_list(self) -> List[str]:
        return [i.lower() for i in self.method_list]

    @property
    def response_model_list(self) -> List[Type[BaseResponseModel]]:
        return self._response_model_list

    def add_response_model_list(self, response_model_list: List[Type[BaseResponseModel]]) -> None:
        for response_model in response_model_list:
            if response_model in self._response_model_list:
                continue
            if issubclass(response_model, PaitResponseModel):
                logging.warning(  # pragma: no cover
                    f"Please replace {self.operation_id}'s response model {response_model}" f" with {BaseResponseModel}"
                )
            self._response_model_list.append(response_model)

    @property
    def extra_openapi_model_list(self) -> List[Type[BaseModel]]:
        return self._extra_openapi_model_list

    @extra_openapi_model_list.setter
    def extra_openapi_model_list(self, item: List[Type[BaseModel]]) -> None:
        self._extra_openapi_model_list.extend(item)

    @property
    def main_plugin(self) -> PluginProtocol:
        return self._main_plugin

    def build_plugin_stack(self) -> None:
        plugin_manager_list: List[PluginManager] = (
            [i for i in self._plugin_list] + [self._param_handler_plugin] + [i for i in self._post_plugin_list]
        )
        self._main_plugin = self.func  # type: ignore
        for plugin_manager in reversed(plugin_manager_list):
            self._main_plugin = plugin_manager.get_plugin(self._main_plugin, self)

    def add_plugin(
        self,
        plugin_list: Optional[List[PluginManager[PrePluginProtocol]]],
        post_plugin_list: Optional[List[PluginManager[PostPluginProtocol]]],
    ) -> None:
        raw_plugin_list: List[PluginManager] = copy.deepcopy(self._plugin_list)
        raw_post_plugin_list: List[PluginManager] = copy.deepcopy(self._post_plugin_list)
        try:
            for plugin_manager in plugin_list or []:
                if issubclass(plugin_manager.plugin_class, PostPluginProtocol):
                    raise ValueError(f"{plugin_manager.plugin_class} is post plugin")
                if not ignore_pre_check:
                    plugin_manager.pre_check_hook(self)
                plugin_manager.pre_load_hook(self)
                self._plugin_list.append(plugin_manager)

            for post_plugin_manager in post_plugin_list or []:
                if issubclass(post_plugin_manager.plugin_class, PrePluginProtocol):
                    raise ValueError(f"{post_plugin_manager.plugin_class} is pre plugin")
                if not ignore_pre_check:
                    post_plugin_manager.pre_check_hook(self)
                post_plugin_manager.pre_load_hook(self)
                self._post_plugin_list.append(post_plugin_manager)
        except Exception as e:
            self._plugin_list = raw_plugin_list
            self._post_plugin_list = raw_post_plugin_list
            raise gen_tip_exc(self.func, RuntimeError(f"{self.func} add plugin error")) from e
        else:
            self.build_plugin_stack()



from typing import Any, Dict, Optional

from any_api.openapi.model.openapi import ExternalDocumentationModel, TagModel

__all__ = ["Tag", "ExternalDocumentationModel", "TagModel"]
_tag_dict: Dict[str, dict] = {}


class Tag(object):
    def __init__(self, name: str, desc: str = "", external_docs: Optional["ExternalDocumentationModel"] = None):
        self._name: str = name
        self._desc: str = desc
        self._external_docs: Optional["ExternalDocumentationModel"] = external_docs
        if self._name in _tag_dict and _tag_dict[self._name] != self.to_dict():
            raise KeyError(f"{self._name} tag already exists")

        _tag_dict[self._name] = self.to_dict()

    @property
    def name(self) -> str:
        return self._name

    @property
    def desc(self) -> str:
        return self._desc

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, Tag):
            return False
        return other.to_dict() == self.to_dict()

    def to_dict(self) -> dict:
        return {"name": self.name, "desc": self.desc, "external_docs": self._external_docs}

    def to_tag_model(self) -> "TagModel":
        try:
            from any_api.openapi.model.openapi import TagModel
        except ImportError:
            raise RuntimeError("please install any-api")
        return TagModel(name=self.name, description=self.desc, externalDocs=self._external_docs)



import inspect
import sys
from contextlib import AbstractContextManager
from typing import Any, Dict, List, Mapping, Optional, Tuple, Type, Union

from pydantic import BaseModel
from typing_extensions import Self  # type: ignore

from pait import field
from pait.exceptions import PaitBaseException
from pait.param_handle.base import BaseParamHandler, raise_multiple_exc
from pait.plugin.base import PluginContext
from pait.util import (
    FuncSig,
    gen_tip_exc,
    get_func_sig,
    get_parameter_list_from_class,
    get_parameter_list_from_pydantic_basemodel,
)


class ParamHandleContext(PluginContext):
    contextmanager_list: List[AbstractContextManager]


class ParamHandler(BaseParamHandler):
    def param_handle(
        self,
        context: "ParamHandleContext",
        _object: Union[FuncSig, Type, None],
        param_list: List["inspect.Parameter"],
        pydantic_model: Optional[Type[BaseModel]] = None,
    ) -> Tuple[List[Any], Dict[str, Any]]:
        args_param_list: List[Any] = []
        kwargs_param_dict: Dict[str, Any] = {}

        for parameter in param_list:
            try:
                if parameter.default != parameter.empty:
                    # kwargs param
                    # support model: def demo(pydantic.BaseModel: BaseModel = pait.field.BaseField())
                    if isinstance(parameter.default, field.Depends):
                        kwargs_param_dict[parameter.name] = self._depend_handle(context, parameter.default.func)
                    else:
                        request_value: Mapping = getattr(
                            context.app_helper.request, parameter.default.get_field_name(), lambda: {}
                        )()
                        self.request_value_handle(parameter, request_value or {}, kwargs_param_dict, pydantic_model)
                else:
                    # args param
                    # support model: model: ModelType
                    self.set_parameter_value_to_args(context, parameter, args_param_list)
            except PaitBaseException as e:
                raise gen_tip_exc(_object, e, parameter, tip_exception_class=self.tip_exception_class)
        return args_param_list, kwargs_param_dict

    def set_parameter_value_to_args(
        self, context: "ParamHandleContext", parameter: inspect.Parameter, func_args: list
    ) -> None:
        """use func_args param faster return and extend func_args"""
        if not self._set_parameter_value_to_args(context, parameter, func_args):
            return
        # support pait_model param(def handle(model: PaitBaseModel))
        _pait_model: Type[BaseModel] = parameter.annotation
        _, kwargs = self.param_handle(
            context,
            None,
            get_parameter_list_from_pydantic_basemodel(_pait_model, context.pait_core_model.default_field_class),
            _pait_model,
        )
        func_args.append(_pait_model(**kwargs))

    def _depend_handle(self, context: "ParamHandleContext", func: Any) -> Any:
        if inspect.isclass(func):
            func = func()
            _, kwargs = self.param_handle(context, func.__class__, get_parameter_list_from_class(func.__class__))
            func.__dict__.update(kwargs)

        func_sig: FuncSig = get_func_sig(func)
        _func_args, _func_kwargs = self.param_handle(context, func_sig, func_sig.param_list)
        func_result: Any = func_sig.func(*_func_args, **_func_kwargs)
        if isinstance(func_result, AbstractContextManager):
            context.contextmanager_list.append(func_result)
            return func_result.__enter__()
        else:
            return func_result

    def _gen_param(self, context: "ParamHandleContext") -> None:
        # check param from pre depend
        for pre_depend in context.pait_core_model.pre_depend_list:
            self._depend_handle(context, pre_depend)

        # gen and check param from func
        func_sig: FuncSig = get_func_sig(context.pait_core_model.func)
        context.args, context.kwargs = self.param_handle(context, func_sig, func_sig.param_list)

        # gen and check param from class
        if context.cbv_param_list:
            _, kwargs = self.param_handle(context, context.cbv_instance.__class__, context.cbv_param_list)
            context.cbv_instance.__dict__.update(kwargs)
        return None

    def __call__(self, context: PluginContext) -> Any:
        error: Optional[Exception] = None
        result: Any = None
        exc_type, exc_val, exc_tb = None, None, None

        param_handle_context: ParamHandleContext = context  # type: ignore
        param_handle_context.contextmanager_list = []

        try:
            self._gen_param(param_handle_context)
            result = super().__call__(context)
        except Exception as e:
            error = e
            exc_type, exc_val, exc_tb = sys.exc_info()
        exc_list: List[Exception] = [error] if error else []
        for contextmanager in param_handle_context.contextmanager_list:
            try:
                contextmanager.__exit__(exc_type, exc_val, exc_tb)
            except Exception as e:
                exc_list.append(e)
        if exc_list:
            raise_multiple_exc(exc_list)
        else:
            return result



from typing import Callable

from starlette.applications import Starlette

from pait.app.base.simple_route import SimpleRoute, add_route_plugin
from pait.app.starlette.plugin.unified_response import UnifiedResponsePlugin


def add_simple_route(
    app: Starlette,
    simple_route: "SimpleRoute",
    prefix: str = "/",
    title: str = "",
    _replace_openapi_url_to_url: Callable[[str], str] = lambda x: x,
) -> None:
    add_route_plugin(simple_route, UnifiedResponsePlugin)
    url: str = prefix + simple_route.url
    url = url.replace("//", "/")
    app.add_route(_replace_openapi_url_to_url(url), simple_route.route, methods=simple_route.methods, name=title)


def add_multi_simple_route(
    app: Starlette,
    *simple_route_list: "SimpleRoute",
    prefix: str = "/",
    title: str = "",
    _replace_openapi_url_to_url: Callable[[str], str] = lambda x: x,
) -> None:
    for simple_route in simple_route_list:
        add_simple_route(app, simple_route, prefix=prefix, title=title)



from dataclasses import MISSING
from typing import Any, Tuple, Type

from sanic.request import Request as _Request

try:
    from sanic.views import CompositionView, HTTPMethodView

    cbv_type_tuple: Tuple[Type[Exception], ...] = (HTTPMethodView, CompositionView)
except ImportError:
    from sanic.views import HTTPMethodView

    cbv_type_tuple = (HTTPMethodView,)
from sanic_testing.testing import SanicTestClient, TestingResponse  # type: ignore

from pait.app.base import BaseAppHelper
from pait.app.sanic._attribute import get_app_attribute
from pait.app.sanic.adapter.request import Request, RequestExtend

__all__ = ["AppHelper", "RequestExtend"]


class AppHelper(BaseAppHelper[_Request, RequestExtend]):
    CbvType = cbv_type_tuple
    app_name = "sanic"

    request_class = Request

    def get_attributes(self, key: str, default: Any = MISSING) -> Any:
        return get_app_attribute(self.raw_request.app, key, default)



from importlib import import_module
from typing import Type

from pait.app.auto_load_app import auto_load_app_class
from pait.app.base.security.http import BaseHTTPBasic, BaseHTTPBearer, BaseHTTPDigest, HTTPBasicCredentials

__all__ = ["HTTPBasic", "HTTPDigest", "HTTPBearer", "HTTPBasicCredentials"]

pait_app_path: str = "pait.app." + auto_load_app_class().__name__.lower() + ".security.http"
HTTPBasic: Type[BaseHTTPBasic] = getattr(import_module(pait_app_path), "HTTPBasic")
HTTPDigest: Type[BaseHTTPDigest] = getattr(import_module(pait_app_path), "HTTPDigest")
HTTPBearer: Type[BaseHTTPBearer] = getattr(import_module(pait_app_path), "HTTPBearer")



from abc import ABCMeta
from typing import Any, Dict, Optional, Type

from pait.model.core import PaitCoreModel
from pait.model.response import BaseResponseModel, FileResponseModel
from pait.plugin.base import PluginContext, PrePluginProtocol
from pait.util import get_pait_response_model


class UnifiedResponsePluginProtocol(PrePluginProtocol):
    response_model_class: Type[BaseResponseModel]

    def _gen_response(self, return_value: Any, context: PluginContext) -> Any:
        raise NotImplementedError

    @classmethod
    def pre_load_hook(cls, pait_core_model: "PaitCoreModel", kwargs: Dict) -> Dict:
        kwargs = super().pre_load_hook(pait_core_model, kwargs)
        if pait_core_model.response_model_list:
            pait_response: Type[BaseResponseModel] = get_pait_response_model(
                pait_core_model.response_model_list,
                target_pait_response_class=kwargs.pop("target_pait_response_class", False),
            )
            if issubclass(pait_response, FileResponseModel):
                raise ValueError(f"Not Support {FileResponseModel.__name__}")
            kwargs["response_model_class"] = pait_response
        else:
            raise ValueError(
                f"The response model list cannot be empty, please add a response model to"
                f" {pait_core_model.func.__qualname__}"
            )
        return kwargs


class UnifiedResponsePlugin(UnifiedResponsePluginProtocol, metaclass=ABCMeta):
    status_code: int
    headers: Optional[dict]
    media_type: Optional[str]

    def __call__(self, context: PluginContext) -> Any:
        if self._is_async_func:
            return self._async_call(context)
        return self._sync_call(context)

    def _sync_call(self, context: PluginContext) -> Any:
        response: Any = super(UnifiedResponsePlugin, self).__call__(context)
        return self._gen_response(response, context)

    async def _async_call(self, context: PluginContext) -> Any:
        response: Any = await super(UnifiedResponsePlugin, self).__call__(context)
        return self._gen_response(response, context)



import inspect
import logging
from typing import TYPE_CHECKING, Any, Callable, Dict, Generic, Type, TypeVar, Union

if TYPE_CHECKING:
    from pait.model.core import PaitCoreModel

from pait.model.context import ContextModel as PluginContext

_PluginT = TypeVar("_PluginT", bound="PluginProtocol")
_NextPluginT = Union[_PluginT, Callable]
logger: logging.Logger = logging.getLogger(__name__)


class PluginProtocol(object):
    def __init__(self, next_plugin: _NextPluginT, pait_core_model: "PaitCoreModel", **kwargs: Any) -> None:
        """Direct init calls are not supported,
        so there is no need to write clearly in init what parameters are needed
        """
        self.next_plugin: _NextPluginT = next_plugin
        self.pait_core_model: "PaitCoreModel" = pait_core_model
        self._is_async_func: bool = inspect.iscoroutinefunction(pait_core_model.func)
        if kwargs:
            for k, v in kwargs.items():
                if getattr(self, k, None) is not None:
                    continue
                setattr(self, k, v)
        self.__post_init__(**kwargs)

    def __post_init__(self, **kwargs: Any) -> None:
        pass

    @classmethod
    def pre_check_hook(cls, pait_core_model: "PaitCoreModel", kwargs: Dict) -> None:
        """The hook that runs the check at startup. If the value of env's PAIT_IGNORE_PRE_CHECK is True,
        it will not be executed.

        Note:
            Failure to execute this stage will cause the plugin to fail to load, but will not affect the use of routes
        """

    @classmethod
    def pre_load_hook(cls, pait_core_model: "PaitCoreModel", kwargs: Dict) -> Dict:
        """Hook for initialization processing, the plugin has not been initialized at this time

        Note:
            Failure to execute this stage will cause the plugin to fail to load, but will not affect the use of routes
        """
        return kwargs  # pragma: no cover

    @classmethod
    def build(cls, **kwargs: Any) -> "PluginManager[_PluginT]":
        """Factory function for generating plugins"""
        return PluginManager(cls, **kwargs)  # type: ignore

    def __call__(self, context: PluginContext) -> Any:
        """The entry function called by the plugin."""
        if isinstance(self.next_plugin, PluginProtocol):
            return self.next_plugin(context)
        else:
            return self.next_plugin(*context.args, **context.kwargs)


class PrePluginProtocol(PluginProtocol):
    """Pre Plugin"""


class PostPluginProtocol(PluginProtocol):
    """Post Plugin"""


class PluginManager(Generic[_PluginT]):
    """
    A proxy for the Pait plugin, Ensure plugins referenced by each route function are isolated

    Note: The plug-in will expose the build method to the user,
        and the user does not need to call it directly `PluginManager`
    """

    def __init__(self, plugin_class: Type[_PluginT], **kwargs: Any):
        self.plugin_class: Type[_PluginT] = plugin_class
        self._kwargs: Any = kwargs

    def pre_check_hook(self, pait_core_model: "PaitCoreModel") -> None:
        self.plugin_class.pre_check_hook(pait_core_model, self._kwargs)

    def pre_load_hook(self, pait_core_model: "PaitCoreModel") -> None:
        self._kwargs = self.plugin_class.pre_load_hook(pait_core_model, self._kwargs)

    def get_plugin(self, next_plugin: _NextPluginT, pait_core_model: "PaitCoreModel") -> _PluginT:
        return self.plugin_class(next_plugin, pait_core_model, **self._kwargs)

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} {self.plugin_class.__name__}>"



import json
from typing import Any, Dict, List, Mapping

from tornado.httputil import HTTPHeaders, HTTPServerRequest

from pait.app.base.adapter.request import BaseRequest, BaseRequestExtend
from pait.util import LazyProperty


class RequestExtend(BaseRequestExtend[HTTPServerRequest]):
    @property
    def scheme(self) -> str:
        return self.request.protocol

    @property
    def path(self) -> str:
        return self.request.path

    @property
    def hostname(self) -> str:
        return self.request.host


class Request(BaseRequest[HTTPServerRequest, RequestExtend]):
    RequestType = HTTPServerRequest
    FormType = dict
    FileType = dict
    HeaderType = dict

    def request_extend(self) -> RequestExtend:
        return RequestExtend(self.request)

    @LazyProperty()
    def body(self) -> dict:
        return json.loads(self.request.body.decode())

    @LazyProperty()
    def cookie(self) -> dict:
        return {i.key: i.value for i in self.request.cookies.values()}

    @LazyProperty()
    def file(self) -> dict:
        return {item["filename"]: item for item in self.request.files["file"]}

    @LazyProperty()
    def form(self) -> dict:
        if self.request.arguments:
            form_dict: dict = {key: value[0].decode() for key, value in self.request.arguments.items()}
        else:
            form_dict = json.loads(self.request.body.decode())  # pragma: no cover
        return form_dict

    def header(self) -> HTTPHeaders:
        return self.request.headers

    def path(self) -> Mapping[str, Any]:
        return self.request_kwargs

    @LazyProperty()
    def query(self) -> dict:
        return {key: value[0].decode() for key, value in self.request.query_arguments.items()}

    @LazyProperty()
    def multiform(self) -> Dict[str, List[Any]]:
        if self.request.arguments:
            return {key: [i.decode() for i in value] for key, value in self.request.arguments.items()}
        else:
            return {key: [value] for key, value in json.loads(self.request.body.decode()).items()}  # pragma: no cover

    @LazyProperty()
    def multiquery(self) -> Dict[str, Any]:
        return {key: [i.decode() for i in value] for key, value in self.request.query_arguments.items()}



import json
import logging
from enum import Enum
from typing import Any, Callable, Dict, Optional, Type
from urllib.parse import urlencode

from any_api.openapi import web_ui
from pydantic import BaseModel, Field

from pait.app.any.util import import_func_from_app
from pait.app.base.simple_route import SimpleRoute
from pait.core import Pait
from pait.field import Depends, Path, Query
from pait.g import config, pait_context
from pait.model import HtmlResponseModel, JsonResponseModel, PaitStatus, Tag, TemplateContext
from pait.openapi.openapi import InfoModel, OpenAPI, ServerModel

logger: logging.Logger = logging.getLogger(__name__)
__all__ = [
    "DocHtmlRespModel",
    "OpenAPIRespModel",
    "AddDocRoute",
    "default_doc_fn_dict",
    "OpenAPI",
    "InfoModel",
    "ServerModel",
    "add_doc_route",
]


class DocHtmlRespModel(HtmlResponseModel):
    class HeaderModel(BaseModel):
        x_example_type: str = Field(default="html", alias="X-Example-Type")

    header: BaseModel = HeaderModel
    description: str = "doc html response"


class OpenAPIRespModel(JsonResponseModel):
    class OpenAPIResponseModel(BaseModel):
        pass

    description: str = "open api json response"
    # TODO replace to openapi response model
    response_data: dict = {}  # type: ignore


default_doc_fn_dict: Dict[str, Callable] = {key.split("_")[1]: getattr(web_ui, key) for key in web_ui.__all__}


class DocEnum(str, Enum):
    pass


class AddDocRoute(object):
    not_found_exc: Exception
    pait: Pait
    add_multi_simple_route: staticmethod

    def __init__(
        self,
        app: Any,
        scheme: Optional[str] = None,
        openapi_json_url_only_path: bool = False,
        prefix: str = "",
        pin_code: str = "",
        title: str = "",
        doc_fn_dict: Optional[Dict[str, Callable]] = None,
        openapi: Optional[Type[OpenAPI]] = None,
        pait: Optional[Pait] = None,
        add_multi_simple_route: Optional[Callable] = None,
        not_found_exc: Optional[Exception] = None,
    ):
        """
        :param app: The app instance to which the doc route is bound
        :param scheme: The scheme of the specified url, if the value is empty,
            it will be automatically selected according to the request

            Note: If the deployment is HTTP and the site configured by Nginx is HTTPS,
                then need to force the specified scheme to be HTTPS
        :param openapi_json_url_only_path: If True, openapi json url does not include HTTP scheme, hostname and port
            Example:
                openapi_json_url_only_path=True,  url: `/api/path`
                openapi_json_url_only_path=False, url: `http://127.0.0.1:8080/api/path`
        :param prefix: Doc route path prefix
        :param pin_code: If pin_code is set, the pin code parameter needs to be added when requesting doc route
            Example:
                pin_code: None, url `http://127.0.0.1:8080/api/path`
                pin_code: 6666, url `http://127.0.0.1:8080/api/path?pin-code=6666`
        :param title: Doc route group name,
            the title of multiple doc routes registered for the same app instance must be different
        :param openapi: OpenAPI class
        :param doc_fn_dict: doc ui dict, default `pait.app.base.doc_route.default_doc_fn_dict`
        :param pait: instance of pait
        :param add_multi_simple_route: add_multi_simple_route
        :param not_found_exc:  not_found_exc
        """
        if pin_code:
            logging.info(f"doc route start pin code:{pin_code}")
        self.scheme: Optional[str] = scheme
        self.openapi_json_url_only_path: bool = openapi_json_url_only_path
        self.prefix: str = prefix or "/"
        self.pin_code: str = pin_code
        self.title: str = title or "Pait Doc"
        self.openapi: Type[OpenAPI] = openapi or OpenAPI
        self._doc_fn_dict: Dict[str, Callable] = doc_fn_dict or default_doc_fn_dict

        # If empty, try to get an available Pait
        _pait: Pait = pait or getattr(self, "pait", None) or import_func_from_app("pait", app=app)  # type: ignore
        self._doc_pait: Pait = _pait.create_sub_pait(
            author=config.author or ("so1n",),
            status=config.status or PaitStatus.release,
            tag=(Tag("pait_doc", desc="pait default doc route"),),
            group="pait_doc",
        )
        self._add_multi_simple_route = (
            add_multi_simple_route
            or getattr(self, "add_multi_simple_route", None)
            or import_func_from_app("add_multi_simple_route", app=app)
        )
        self._not_found_exc = (
            not_found_exc
            or getattr(self, "not_found_exc", None)
            or import_func_from_app("http_exception", app=app)(status_code=404, message="Not Found")
        )
        self._gen_route(app)

    def _get_request_pin_code(self, pin_code: str = Query.i("", alias="pin-code")) -> Optional[str]:
        if self.pin_code and pin_code != self.pin_code:
            raise self._not_found_exc
        return pin_code

    @staticmethod
    def _get_request_template_map(extra_key: bool = False) -> Callable:
        def _get_request_template_map(
            url_dict: Dict[str, Any] = Query.i(
                default_factory=dict,
                raw_return=True,
                description="Set the template variable, for example, there is a template parameter token, "
                "then the requested parameter is template-token=xxx",
            )
        ) -> dict:
            template_dict: Dict[str, Any] = {}
            for k, v in url_dict.items():
                if k.startswith("template-"):
                    if extra_key:
                        k = k.replace("template-", "")
                    template_dict[k] = v
            return template_dict

        return _get_request_template_map

    def _gen_url_fn(self) -> Callable:
        def _get_openapi_json_url(
            r_pin_code: str = Depends.i(self._get_request_pin_code),
            url_dict: Dict[str, Any] = Depends.i(self._get_request_template_map()),
        ) -> str:
            re = pait_context.get().app_helper.request.request_extend()
            _scheme: str = self.scheme or re.scheme
            if self.openapi_json_url_only_path:
                openapi_json_url: str = f"{'/'.join(re.path.split('/')[:-1])}/openapi.json"
            else:
                openapi_json_url = f"{_scheme}://{re.hostname}{'/'.join(re.path.split('/')[:-1])}/openapi.json"
            if r_pin_code:
                url_dict["pin-code"] = r_pin_code
            openapi_json_url += "?" + urlencode(url_dict)
            return openapi_json_url

        return _get_openapi_json_url

    def _gen_route(self, app: Any) -> Any:
        self._add_multi_simple_route(
            app,
            SimpleRoute(url="/openapi.json", route=self._get_openapi_route(app), methods=["GET"]),
            SimpleRoute(
                url="/{route_path}",
                route=self._get_doc_route(),
                methods=["GET"],
            ),
            prefix=self.prefix,
            title=self.title,
        )

    def _get_doc_route(self) -> Callable:
        dynamic_enum_class: Type[Enum] = DocEnum(  # type: ignore
            "DynamicEnum", {key: key for key in self._doc_fn_dict.keys()}
        )

        @self._doc_pait(
            pre_depend_list=[self._get_request_pin_code],
            response_model_list=[DocHtmlRespModel],
            feature_code=self.title,
        )
        def _doc_route(
            route_path: dynamic_enum_class = Path.i(description="doc ui html"),  # type: ignore
            url: str = Depends.i(self._gen_url_fn()),
        ) -> str:
            get_doc_route: Callable = self._doc_fn_dict[route_path.value]  # type: ignore
            return get_doc_route(url, title=self.title)

        _doc_route.__name__ = self.title + "doc_route"
        _doc_route.__qualname__ = _doc_route.__qualname__.replace("._doc_route", "." + _doc_route.__name__)
        return _doc_route

    def _get_openapi_route(self, app: Any) -> Callable:
        @self._doc_pait(
            pre_depend_list=[self._get_request_pin_code],
            response_model_list=[OpenAPIRespModel],
            feature_code=self.title,
        )
        def _openapi_route(
            url_dict: Dict[str, Any] = Depends.i(self._get_request_template_map(extra_key=True)),
        ) -> dict:
            re = pait_context.get().app_helper.request.request_extend()
            _scheme: str = self.scheme or re.scheme
            with TemplateContext(url_dict):
                pait_openapi: OpenAPI = self.openapi(app)
                pait_openapi.model.info.title = self.title
                pait_openapi.model.servers.insert(0, ServerModel(url=f"{_scheme}://{re.hostname}"))
                return json.loads(pait_openapi.content())

        return _openapi_route


def add_doc_route(
    app: Any,
    scheme: Optional[str] = None,
    openapi_json_url_only_path: bool = False,
    prefix: str = "",
    pin_code: str = "",
    title: str = "",
    doc_fn_dict: Optional[Dict[str, Callable]] = None,
    openapi: Optional[Type[OpenAPI]] = None,
    pait: Optional[Pait] = None,
    add_multi_simple_route: Optional[Callable] = None,
    not_found_exc: Optional[Exception] = None,
) -> None:
    AddDocRoute(
        scheme=scheme,
        openapi_json_url_only_path=openapi_json_url_only_path,
        prefix=prefix,
        pin_code=pin_code,
        title=title,
        app=app,
        openapi=openapi,
        doc_fn_dict=doc_fn_dict,
        pait=pait,
        add_multi_simple_route=add_multi_simple_route,
        not_found_exc=not_found_exc,
    )



from importlib import import_module

from pait.app.auto_load_app import auto_load_app_class
from pait.plugin.unified_response import UnifiedResponsePlugin as _UnifiedResponsePlugin

__all__ = ["UnifiedResponsePlugin"]

pait_app_path: str = "pait.app." + auto_load_app_class().__name__.lower()
UnifiedResponsePlugin: "_UnifiedResponsePlugin" = getattr(
    import_module(pait_app_path + ".plugin"), "UnifiedResponsePlugin"
)



from pait.app.any import (
    Pait,
    SimpleRoute,
    add_multi_simple_route,
    add_simple_route,
    get_app_attribute,
    http_exception,
    load_app,
    load_class_app,
    pait,
    set_app_attribute,
)



import json
from typing import Any

from tornado.web import RequestHandler

from pait.plugin.base import PluginContext
from pait.plugin.check_json_resp import CheckJsonRespPlugin as _CheckJsonRespPlugin

__all__ = ["CheckJsonRespPlugin"]


class CheckJsonRespPlugin(_CheckJsonRespPlugin):
    @staticmethod
    def get_json(response_data: Any, context: PluginContext) -> dict:
        if response_data is None:
            tornado_handle: RequestHandler = context.args[0]
            if "application/json; charset=UTF-8" in tornado_handle._headers.get_list("Content-Type"):
                return json.loads(tornado_handle._write_buffer[-1].decode())
            else:
                raise ValueError(
                    f"Expected 'application/json', but got '{tornado_handle._headers.get_list('Content-Type')}'"
                )
        else:
            raise TypeError(f"Expected type must None but got type {type(response_data)}")

    async def _async_call(self, context: PluginContext) -> Any:
        try:
            return await super()._async_call(context)
        except Exception:
            tornado_handle: RequestHandler = context.args[0]
            tornado_handle.clear()
            raise



from typing import Any

from flask import Response

from pait.plugin.base import PluginContext
from pait.plugin.check_json_resp import CheckJsonRespPlugin as _CheckJsonRespPlugin

__all__ = ["CheckJsonRespPlugin"]


class CheckJsonRespPlugin(_CheckJsonRespPlugin):
    @staticmethod
    def get_json(response_data: Any, context: PluginContext) -> dict:
        if isinstance(response_data, Response):
            return response_data.get_json()
        else:
            raise TypeError(f"Expected type must {Response} but got type {type(response_data)}")



import copy
import difflib
from typing import TYPE_CHECKING, Any, Callable, Dict, Generic, List, Mapping, Optional, Type, TypeVar
from urllib.parse import urlencode

from pydantic import BaseModel, ValidationError

from pait.model import response
from pait.util import gen_example_dict_from_schema, gen_example_value_from_python, get_pait_response_model

if TYPE_CHECKING:
    from pait.model import PaitCoreModel

RESP_T = TypeVar("RESP_T")
_error_separator = "\n" + ">" * 20 + "\n"


class CheckResponseException(Exception):
    def __init__(
        self,
        status_code: Optional[int] = None,
        media_type: Optional[str] = None,
        headers: Optional[Mapping] = None,
        resp: Any = None,
        func: Optional[Callable] = None,
        message: Optional[str] = None,
    ):
        self.status_code: Optional[int] = status_code
        self.media_type: Optional[str] = media_type
        self.headers: Optional[Mapping] = headers
        self.resp: Any = resp
        self.func: Optional[Callable] = func
        self.message: Optional[str] = message
        super().__init__(self.message)


class BaseTestHelper(Generic[RESP_T]):
    client: Any

    def __init__(
        self,
        client: Any,
        func: Callable,
        pait_dict: Optional[Dict[str, "PaitCoreModel"]] = None,
        body_dict: Optional[dict] = None,
        cookie_dict: Optional[dict] = None,
        file_dict: Optional[dict] = None,
        form_dict: Optional[dict] = None,
        header_dict: Optional[dict] = None,
        path_dict: Optional[dict] = None,
        query_dict: Optional[dict] = None,
        strict_inspection_check_json_content: bool = True,
        enable_assert_response: bool = True,
        target_pait_response_class: Optional[Type["response.BaseResponseModel"]] = None,
    ):
        """
        :param client:  test client
        :param func:  route handle decorated by pait
        :param pait_dict:  pait core data
        :param body_dict:  request body param
        :param cookie_dict:  request cookie param
        :param file_dict:  request file param
        :param form_dict:  request form param
        :param header_dict:  request herder param
        :param path_dict:  request path param
        :param query_dict:  request query param
        """
        pait_id: str = getattr(func, "_pait_id", "")
        if not pait_id:
            raise RuntimeError(f"Can not found pait id from {func}")

        self.client: Any = client
        self.func: Callable = func
        # pait dict handle
        if pait_dict:
            self.pait_dict: Dict[str, "PaitCoreModel"] = pait_dict
        else:
            self.pait_dict = self._gen_pait_dict()
        self.pait_core_model: "PaitCoreModel" = self.pait_dict[pait_id]

        self.body_dict: Optional[dict] = body_dict
        self.cookie_dict: Optional[dict] = cookie_dict
        self.file_dict: Optional[dict] = file_dict
        self.form_dict: Optional[dict] = form_dict
        self.header_dict: dict = header_dict or {}
        self.path_dict: Optional[dict] = path_dict
        self.query_dict: Optional[dict] = query_dict
        self.strict_inspection_check_json_content: bool = strict_inspection_check_json_content
        self.enable_assert_response: bool = enable_assert_response

        # path handle
        self.path: str = self.pait_core_model.path
        if self.path_dict:
            path_list: List[str] = self.path.split("/")
            new_path_list: List[str] = []
            for sub_path in path_list:
                if not sub_path:
                    continue
                new_sub_path: Optional[str] = self._replace_path(sub_path)
                if new_sub_path:
                    new_path_list.append(new_sub_path)
                else:
                    new_path_list.append(sub_path)
            self.path = "/".join([str(i) for i in new_path_list])

        # add query path by path
        if self.query_dict:
            self.path = self.path + "?" + urlencode(self.query_dict, True)
        if not self.path.startswith("/"):
            self.path = "/" + self.path

        self.target_pait_response_class: Optional[Type["response.BaseResponseModel"]] = target_pait_response_class
        self._app_init_field()

    def _app_init_field(self) -> None:
        """init request param by application framework"""
        raise NotImplementedError()

    def _gen_pait_dict(self) -> Dict[str, "PaitCoreModel"]:
        """load pait dict"""
        raise NotImplementedError()

    @staticmethod
    def _get_status_code(resp: RESP_T) -> int:
        """get response status code"""
        raise NotImplementedError()

    @staticmethod
    def _get_content_type(resp: RESP_T) -> str:
        """get response content type"""
        raise NotImplementedError()

    @staticmethod
    def _get_json(resp: RESP_T) -> dict:
        """get json response"""
        raise NotImplementedError()

    @staticmethod
    def _get_headers(resp: RESP_T) -> Mapping:
        """get response header"""
        raise NotImplementedError()

    @staticmethod
    def _get_text(resp: RESP_T) -> str:
        """get text response"""
        raise NotImplementedError()

    @staticmethod
    def _get_bytes(resp: RESP_T) -> bytes:
        """get bytes response"""
        raise NotImplementedError()

    def _check_diff_resp_dict(
        self, raw_resp: Any, default_resp: Any, parent_key_list: Optional[List[str]] = None
    ) -> bool:
        """check resp data structure diff, Only return True to pass the verification"""
        raw_parent_key_list: List[str] = parent_key_list or []
        try:
            if isinstance(raw_resp, dict):
                for key, value in raw_resp.items():
                    parent_key_list = copy.deepcopy(raw_parent_key_list)
                    parent_key_list.append(key)
                    if not self._check_diff_resp_dict(value, default_resp[key], parent_key_list):
                        return False
                return True
            elif isinstance(raw_resp, list) or isinstance(raw_resp, tuple):
                if (
                    raw_resp
                    and default_resp
                    and not self._check_diff_resp_dict(raw_resp[0], default_resp[0], parent_key_list)
                ):
                    return False
                return True
            else:
                return True
        except KeyError:
            raise RuntimeError(f"Can not found key from model, key:{' -> '.join(parent_key_list or '')}")

    def _assert_response(self, resp: RESP_T) -> Optional[Exception]:
        """Whether the structure of the check response is correct"""
        if not self.pait_core_model.response_model_list:
            return None

        real_response_model: Optional[Type[response.BaseResponseModel]] = None
        max_quick_ratio: float = 0.0
        model_check_msg_dict: Dict[Type[response.BaseResponseModel], List[str]] = {}
        response_model_list: List[Type[response.BaseResponseModel]] = (
            self.pait_core_model.response_model_list
            if not self.target_pait_response_class
            else [
                get_pait_response_model(
                    self.pait_core_model.response_model_list,
                    target_pait_response_class=self.target_pait_response_class,
                )
            ]
        )
        for response_model in response_model_list:
            error_msg_list: List[str] = []
            status_code: int = self._get_status_code(resp)
            content_type: str = self._get_content_type(resp)
            if status_code not in response_model.status_code:
                error_msg_list.append(
                    "check status code error. "
                    f"Get the response with a status code of {status_code}, "
                    f"but the response_model specifies a status of {response_model.status_code}"
                )

            if response_model.media_type not in content_type:
                error_msg_list.append(
                    "check media type error."
                    f"Get the response with a media type of {content_type}, "
                    f"but the response_model specifies a media type of {response_model.media_type}"
                )

            resp_header: Mapping = self._get_headers(resp)
            for key, _ in response_model.get_header_example_dict().items():
                if key not in resp_header:
                    error_msg_list.append(
                        f"check header error. Can not found header:{key} in {str(resp_header).strip()}"
                    )

            # check content
            if issubclass(response_model, response.JsonResponseModel):
                response_data_model: Type[BaseModel] = response_model.response_data
                resp_dict: Optional[dict] = None
                try:
                    resp_dict = self._get_json(resp)
                except Exception:
                    pass

                if not resp_dict and not isinstance(resp_dict, dict):
                    continue
                response_data_default_dict: dict = gen_example_dict_from_schema(response_data_model.schema())
                ratio: float = difflib.SequenceMatcher(
                    None,
                    str(gen_example_value_from_python(resp_dict)),
                    str(response_data_default_dict),
                ).quick_ratio()
                if ratio > max_quick_ratio:
                    max_quick_ratio = ratio
                    real_response_model = response_model
                try:
                    response_data_model(**resp_dict)
                    # like Basemodel.config.extra = forbid
                    if self.strict_inspection_check_json_content and not self._check_diff_resp_dict(
                        resp_dict, response_data_default_dict
                    ):
                        differ = difflib.Differ()
                        diff_iterator = differ.compare(str(response_data_default_dict), str(resp_dict))
                        diff_content = "\n".join(diff_iterator)
                        error_msg_list.append(f"check json structure error, \n{diff_content}")
                except (ValidationError, RuntimeError) as e:
                    error_msg_list.append(f"check json content error, exec: {e}")

            elif issubclass(response_model, response.HtmlResponseModel) or issubclass(
                response_model, response.TextResponseModel
            ):
                real_response_model = response_model
                if not isinstance(self._get_text(resp), type(response_model.response_data)):
                    error_msg_list.append("check text content type error")  # pragma: no cover
            elif issubclass(response_model, response.FileResponseModel):
                real_response_model = response_model
                if not isinstance(self._get_bytes(resp), type(response_model.response_data)):
                    error_msg_list.append("check bytes content type error")  # pragma: no cover
            else:
                return TypeError(f"Pait not support response model:{response_model}")
            if not error_msg_list:
                return None
            else:
                model_check_msg_dict[response_model] = error_msg_list
        if real_response_model in model_check_msg_dict:
            error_str: str = "\n    ".join(model_check_msg_dict[real_response_model])
            return CheckResponseException(
                status_code=self._get_status_code(resp),
                media_type=self._get_content_type(resp),
                headers=self._get_headers(resp),
                resp=resp,
                func=self.func,
                message=(
                    f"maybe error result: {_error_separator}{error_str}\n"
                    f"{_error_separator}by response model:{real_response_model}\n"
                ),
            )

        error_str = ""
        for k, v in model_check_msg_dict.items():
            error_str += str(k) + ":\n" + "    \n".join(v)
        return CheckResponseException(
            status_code=self._get_status_code(resp),
            media_type=self._get_content_type(resp),
            headers=self._get_headers(resp),
            resp=resp,
            func=self.func,
            message=(
                f"response error result: {_error_separator}{error_str} \n"
                f"{_error_separator}by response model list:{self.pait_core_model.response_model_list}.\n"
            ),
        )

    def _replace_path(self, path_str: str) -> Optional[str]:
        raise NotImplementedError()

    def _real_request(self, method: str) -> RESP_T:
        """Whether the structure of the check response is correct"""
        raise NotImplementedError()

    def request(self, method: Optional[str] = None) -> RESP_T:
        """user call test request api"""
        if not method:
            # auto select method
            if len(self.pait_core_model.method_list) == 1:
                method = self.pait_core_model.method_list[0]
            else:
                raise RuntimeError(
                    f"Pait Can not auto select method, please choice method in {self.pait_core_model.method_list}"
                )
        resp: RESP_T = self._real_request(method)
        if self.pait_core_model.response_model_list and self.enable_assert_response:
            exc = self._assert_response(resp)
            if exc:
                raise exc
        return resp

    def json(self, method: Optional[str] = None) -> dict:
        return self._get_json(self.request(method))

    def text(self, method: Optional[str] = None) -> str:
        return self._get_text(self.request(method))

    def get(self) -> RESP_T:
        return self.request("GET")

    def patch(self) -> RESP_T:
        return self.request("PATCH")

    def post(self) -> RESP_T:
        return self.request("POST")

    def head(self) -> RESP_T:
        return self.request("HEAD")

    def put(self) -> RESP_T:
        return self.request("PUT")

    def delete(self) -> RESP_T:
        return self.request("DELETE")

    def options(self) -> RESP_T:
        return self.request("OPTIONS")



from dataclasses import MISSING
from typing import TYPE_CHECKING, Any, Callable, List, Mapping, Optional, TypeVar

from pydantic import BaseModel
from pydantic.fields import FieldInfo, Undefined
from pydantic.typing import NoArgAnyCallable

from pait.types import CallType, ParamSpec

if TYPE_CHECKING:
    from pait.openapi.openapi import LinksModel


_T = TypeVar("_T")


class ExtraParam(BaseModel):
    pass


class BaseField(FieldInfo):
    field_name: str = ""
    media_type: str = "*/*"
    openapi_serialization: Optional[dict] = None

    def __init__(
        self,
        default: Any = Undefined,
        *,
        links: "Optional[LinksModel]" = None,
        media_type: str = "",
        openapi_serialization: Any = None,
        raw_return: bool = False,
        example: Any = MISSING,
        not_value_exception: Optional[Exception] = None,
        openapi_include: bool = True,
        extra_param_list: Optional[List[ExtraParam]] = None,
        # pydantic.Field param
        default_factory: Optional[NoArgAnyCallable] = None,
        alias: Optional[str] = None,
        title: Optional[str] = None,
        description: Optional[str] = None,
        const: Optional[bool] = None,
        gt: Optional[float] = None,
        ge: Optional[float] = None,
        lt: Optional[float] = None,
        le: Optional[float] = None,
        multiple_of: Optional[float] = None,
        min_items: Optional[int] = None,
        max_items: Optional[int] = None,
        min_length: Optional[int] = None,
        max_length: Optional[int] = None,
        regex: Optional[str] = None,
        **extra: Any,
    ):
        """
        :param links: OpenAPI Link model
        :param media_type: default media type of request body
        :param openapi_serialization: Additional description of OpenAPI Schema
        :param raw_return: If True, the return value will be returned as is (the request key will be invalidated).
        :param example: example data
        :param not_value_exception:
            The exception thrown when the corresponding value cannot be obtained from the request,
            default value `pait.exceptions.NotFoundValueException`
        :param openapi_include: Whether it is used by the OpenAPI
        :param extra_param_list: Extended parameters for plug-ins to use
        """
        # Same checks as pydantic, checked in advance here
        if default is not Undefined and default_factory is not None:
            raise ValueError("cannot specify both default and default_factory")  # pragma: no cover

        # Reduce runtime judgments by preloading
        if default is not Undefined:
            self.request_value_handle = self.request_value_handle_by_default  # type: ignore
        elif default_factory is not None:
            self.request_value_handle = self.request_value_handle_by_default_factory  # type: ignore

        if getattr(example, "__class__", None) is not MISSING.__class__:
            extra["example"] = example
        if links:
            extra["links"] = links

        #######################################################
        # These parameters will not be used in pydantic.Field #
        #######################################################
        self.openapi_include: bool = openapi_include
        self.raw_return = raw_return
        self.not_value_exception: Optional[Exception] = not_value_exception
        self.media_type = media_type or self.__class__.media_type
        # if not alias, pait will set the key name to request_key in the preload phase
        self.request_key: str = alias or ""
        self.openapi_serialization = openapi_serialization or self.__class__.openapi_serialization
        self.extra_param_list: List[ExtraParam] = extra_param_list or []

        super().__init__(
            default,
            default_factory=default_factory,
            alias=alias,
            title=title,
            description=description,
            const=const,
            gt=gt,
            ge=ge,
            lt=lt,
            le=le,
            multiple_of=multiple_of,
            min_items=min_items,
            max_items=max_items,
            min_length=min_length,
            max_length=max_length,
            regex=regex,
            **extra,
        )

    def set_alias(self, value: Optional[str]) -> None:
        self.alias = value
        if value is not None:
            self.request_key = value

    def request_value_handle(self, request_value: Mapping) -> Any:
        return request_value.get(self.request_key, Undefined)

    def request_value_handle_by_default(self, request_value: Mapping) -> Any:
        return request_value.get(self.request_key, self.default)

    def request_value_handle_by_default_factory(self, request_value: Mapping) -> Any:
        return request_value.get(self.request_key, self.default_factory())

    @property
    def links(self) -> "Optional[LinksModel]":
        return self.extra.get("links", None)

    @classmethod
    def get_field_name(cls) -> str:
        return cls.field_name or cls.__name__.lower()

    @classmethod
    def i(
        cls,
        default: Any = Undefined,
        *,
        links: "Optional[LinksModel]" = None,
        raw_return: bool = False,
        media_type: str = "",
        example: Any = MISSING,
        openapi_serialization: Any = None,
        not_value_exception: Optional[Exception] = None,
        openapi_include: bool = True,
        default_factory: Optional[NoArgAnyCallable] = None,
        alias: Optional[str] = None,
        title: Optional[str] = None,
        description: Optional[str] = None,
        const: Optional[bool] = None,
        gt: Optional[float] = None,
        ge: Optional[float] = None,
        lt: Optional[float] = None,
        le: Optional[float] = None,
        multiple_of: Optional[float] = None,
        min_items: Optional[int] = None,
        max_items: Optional[int] = None,
        min_length: Optional[int] = None,
        max_length: Optional[int] = None,
        regex: Optional[str] = None,
        extra_param_list: Optional[List[ExtraParam]] = None,
        **extra: Any,
    ) -> Any:
        """ignore mypy tip"""
        return cls(
            default,
            raw_return=raw_return,
            links=links,
            example=example,
            media_type=media_type,
            openapi_serialization=openapi_serialization,
            not_value_exception=not_value_exception,
            openapi_include=openapi_include,
            default_factory=default_factory,
            alias=alias,
            title=title,
            description=description,
            const=const,
            gt=gt,
            ge=ge,
            lt=lt,
            le=le,
            multiple_of=multiple_of,
            min_items=min_items,
            max_items=max_items,
            min_length=min_length,
            max_length=max_length,
            regex=regex,
            extra_param_list=extra_param_list,
            **extra,
        )

    @classmethod
    def t(
        cls,
        default: _T = Undefined,
        *,
        links: "Optional[LinksModel]" = None,
        raw_return: bool = False,
        media_type: str = "",
        example: _T = MISSING,  # type: ignore
        openapi_serialization: Any = None,
        not_value_exception: Optional[Exception] = None,
        openapi_include: bool = True,
        default_factory: Optional[Callable[[], _T]] = None,
        alias: Optional[str] = None,
        title: Optional[str] = None,
        description: Optional[str] = None,
        const: Optional[bool] = None,
        gt: Optional[float] = None,
        ge: Optional[float] = None,
        lt: Optional[float] = None,
        le: Optional[float] = None,
        multiple_of: Optional[float] = None,
        min_items: Optional[int] = None,
        max_items: Optional[int] = None,
        min_length: Optional[int] = None,
        max_length: Optional[int] = None,
        regex: Optional[str] = None,
        extra_param_list: Optional[List[ExtraParam]] = None,
        **extra: Any,
    ) -> _T:
        """Limited type hint support"""
        return cls(
            default,
            raw_return=raw_return,
            links=links,
            example=example,
            media_type=media_type,
            openapi_serialization=openapi_serialization,
            not_value_exception=not_value_exception,
            openapi_include=openapi_include,
            default_factory=default_factory,
            alias=alias,
            title=title,
            description=description,
            const=const,
            gt=gt,
            ge=ge,
            lt=lt,
            le=le,
            multiple_of=multiple_of,
            min_items=min_items,
            max_items=max_items,
            min_length=min_length,
            max_length=max_length,
            regex=regex,
            extra_param_list=extra_param_list,
            **extra,
        )

    @classmethod
    def from_pydantic_field(cls, field: FieldInfo) -> "BaseField":
        """use replace pydantic property field"""
        return cls(
            field.default,
            default_factory=field.default_factory,
            alias=field.alias,
            title=field.title,
            description=field.description,
            const=field.const,
            gt=field.gt,
            ge=field.ge,
            lt=field.lt,
            le=field.le,
            multiple_of=field.multiple_of,
            min_items=field.min_items,
            max_items=field.max_items,
            min_length=field.min_length,
            max_length=field.max_length,
            regex=field.regex,
            extra_param_list=field.extra.pop("extra_param_list", None),
            **field.extra,
        )


class Json(BaseField):
    media_type: str = "application/json"
    field_name = "body"


class Body(Json):
    """Compatible with the old interface, it may be removed in the future"""


class Cookie(BaseField):
    pass


class File(BaseField):
    media_type: str = "multipart/form-data"


class Form(BaseField):
    media_type: str = "application/x-www-form-urlencoded"


class MultiForm(BaseField):
    media_type: str = "application/x-www-form-urlencoded"
    openapi_serialization: Optional[dict] = {"style": "form", "explode": True}


class Header(BaseField):
    pass


class Path(BaseField):
    pass


class Query(BaseField):
    pass


class MultiQuery(BaseField):
    pass


P = ParamSpec("P")
R_T = TypeVar("R_T")


class Depends(object):
    def __init__(self, func: CallType):
        self.func: CallType = func

    @classmethod
    def i(cls, func: CallType) -> Any:
        return cls(func)

    @classmethod
    def t(cls, func: Callable[P, R_T]) -> R_T:  # type: ignore
        return cls(func)  # type: ignore


def is_pait_field(pait_field: Any) -> bool:
    return isinstance(pait_field, (BaseField, Depends))


def is_pait_field_class(pait_field: Any) -> bool:
    return issubclass(pait_field, (BaseField, Depends))



from typing import Any

from pait.app.flask.adapter.response import gen_response
from pait.plugin.unified_response import PluginContext
from pait.plugin.unified_response import UnifiedResponsePlugin as BaseUnifiedResponsePlugin
from pait.plugin.unified_response import UnifiedResponsePluginProtocol as BaseUnifiedResponsePluginProtocol


def _gen_response(self: BaseUnifiedResponsePluginProtocol, return_value: Any, context: PluginContext) -> Any:
    return gen_response(return_value, self.response_model_class, *context.args, **context.kwargs)


class UnifiedResponsePlugin(BaseUnifiedResponsePlugin):
    _gen_response = _gen_response


class UnifiedResponsePluginProtocol(BaseUnifiedResponsePluginProtocol):
    _gen_response = _gen_response



from dataclasses import MISSING
from typing import Any, List

from flask import Request as _Request
from flask import current_app, request
from flask.views import View

from pait.app.base import BaseAppHelper
from pait.app.flask._attribute import get_app_attribute
from pait.app.flask.adapter.request import Request, RequestExtend

__all__ = ["AppHelper", "RequestExtend"]


class AppHelper(BaseAppHelper[_Request, RequestExtend]):
    CbvType = (View,)
    app_name = "flask"

    request_class = Request

    def _get_request(self, args: List[Any]) -> _Request:
        return request

    def get_attributes(self, key: str, default: Any = MISSING) -> Any:
        return get_app_attribute(current_app, key, default)



from dataclasses import MISSING
from typing import Any, Generic, List, Mapping, Optional, Type

from pait.app.base.adapter.request import BaseRequest, BaseRequestExtend, RequestExtendT, RequestT

__all__ = ["BaseRequest", "RequestT", "RequestExtendT", "BaseAppHelper", "BaseRequestExtend"]


class BaseAppHelper(Generic[RequestT, RequestExtendT]):
    """Provide a unified framework call interface for pait"""

    # The class that defines the request object corresponding to the framework (consistent with Request T)
    CbvType: tuple = (Type,)  # The class that defines the cbv object corresponding to the framework
    app_name: str = "BaseAppHelper"  # Define the name corresponding to the framework

    request_class: Type[BaseRequest[RequestT, RequestExtendT]] = BaseRequest

    def __init__(self, args: List[Any], kwargs: Mapping[str, Any]):
        """
        Extract the required data from the passed parameters,
        such as the self parameter in cvb mode, the request parameter in starletter
        """
        if args and isinstance(args[0], self.CbvType):
            self.cbv_instance: Any = args[0]
        else:
            self.cbv_instance = None

        request: Optional[RequestT] = self._get_request(args)
        if request is None:
            raise ValueError("Unable to get request object")
        self.raw_request: RequestT = request
        self.request: BaseRequest[RequestT, RequestExtendT] = self.request_class(request, args, kwargs)

    def _get_request(self, args: List[Any]) -> Optional[RequestT]:
        for param in args:
            if isinstance(param, self.request_class.RequestType):  # type: ignore
                return param
        return None

    def get_attributes(self, key: str, default: Any = MISSING) -> Any:
        raise NotImplementedError
